from __future__ import annotations

import maria
import pytest


def test_cmb_patch():
    plan = maria.Plan.generate(
        start_time="2025-01-01",
        scan_pattern="daisy",
        scan_options={"radius": 5, "speed": 1},  # in degrees
        duration=120,  # in seconds
        sample_rate=50,  # in Hz
        scan_center=(0, -50),
        jitter=0,
        frame="ra/dec",
    )

    cmb_patch = maria.cmb.generate_cmb_patch(width=15, center=plan.center())

    sim = maria.Simulation(
        instrument="test/1deg",
        plans=plan,
        site="cerro_toco",
        map=cmb_patch,
        noise=False,
    )

    tod = sim.run()[0].to("K_CMB")

    # should be around 110 uK
    cmb_rms_K = tod.signal.std(axis=-1).compute()

    assert all(cmb_rms_K > 0.5 * cmb_patch.rms)
    assert all(cmb_rms_K < 2.0 * cmb_patch.rms)



from __future__ import annotations

import maria
import matplotlib.pyplot as plt
import numpy as np
import pytest

plt.close("all")


arrays1 = {
    "subarray-1": {
        "n": 500,
        "primary_size": 10,
        "field_of_view": 2,
        "shape": "hexagon",
        "bands": [{"center": 30e9, "width": 5e9}, {"center": 40e9, "width": 5e9}],
    },
    "subarray-2": {
        "n": 500,
        "primary_size": 10,
        "field_of_view": 2,
        "shape": "hexagon",
        "bands": [{"center": 90e9, "width": 5e9}, {"center": 150e9, "width": 5e9}],
        "rotation": 90,
    },
}

# /Users/tom/maria/src/maria/array/

arrays2 = [{"file": "alma/alma_f144.csv", "primary_size": 12, "bands": ["alma/f144"]}]


offsets = np.random.uniform(low=0, high=1, size=(100, 2))
arrays3 = [
    {"bands": ["act/pa4/f220"], "primary_size": 5, "sky_x": offsets[..., 0], "sky_y": offsets[..., 1], "degrees": True}
]

array_configs_to_test = [arrays1, arrays2, arrays3]


@pytest.mark.parametrize("array_config", array_configs_to_test)
def test_generate_array(array_config):
    instrument = maria.Instrument(arrays=array_config)



from __future__ import annotations

import dask.array as da
import numpy as np


def extrude(
    values: np.array,
    A: np.array,
    B: np.array,
    n_steps: int,
    n_i: int,
    n_j: int,
    i_sample_index: int,
    j_sample_index: int,
):
    # muy rapido
    BUFFER = np.zeros((n_steps + n_i) * n_j)
    BUFFER = da.random.random(size=((n_steps + n_i) * n_j))
    BUFFER[n_steps * n_j :] = values

    # remember that (e, c) -> n_c * e + c
    for buffer_index in np.arange(n_steps)[::-1]:
        BUFFER[buffer_index * n_j + np.arange(n_j)] = A @ BUFFER[
            n_j * (buffer_index + 1 + i_sample_index) + j_sample_index
        ] + B @ np.random.standard_normal(size=n_j)

    return BUFFER[: n_steps * n_j]


DEFAULT_ATMOSPHERE_CONFIG = {
    "min_depth": 500,
    "max_depth": 3000,
    "n_layers": 4,
    "min_beam_res": 4,
}


MIN_SAMPLES_PER_RIBBON = 2
JITTER_LEVEL = 1e-4



try:
    import torch

    torch.set_num_threads(1)
except ImportError:
    raise ImportError("To do maximum likelihood mapping in maria, please install PyTorch (pip install torch)")

from collections.abc import Sequence

import numpy as np
import scipy as sp
from tqdm import tqdm, trange

from ..map import ProjectionMap
from ..tod import TOD
from ..utils import decompose
from .base import BaseProjectionMapper
from .bin_mapper import BinMapper


def remove_ends(data, ntip=16):
    start_values = data[..., :ntip].mean(axis=-1)
    end_values = data[..., -ntip:].mean(axis=-1)

    return data - np.linspace(start_values, end_values, data.shape[-1]).T


class MaximumLikelihoodSolution(torch.nn.Module):
    def __init__(self, data: torch.Tensor, mask: torch.Tensor, dtype: type = torch.float):
        super(MaximumLikelihoodSolution, self).__init__()

        data_std = data[mask].detach().std()
        data_std = data_std if data_std > 0 else torch.tensor(1.0)
        self.log_scale = torch.nn.Parameter(data_std.log())
        self.unscaled_x = torch.nn.Parameter(data[mask] / data_std)
        self.mask = mask

    def x(self):
        x = torch.zeros_like(self.mask, dtype=torch.float)
        x[self.mask] = self.log_scale.exp() * self.unscaled_x
        return x


class MaximumLikelihoodMapper(BaseProjectionMapper):
    def __init__(
        self,
        tods: Sequence[TOD],
        center: tuple[float, float] = None,
        stokes: str = None,
        width: float = None,
        height: float = None,
        resolution: float = None,
        frame: str = "ra/dec",
        units: str = "K_RJ",
        degrees: bool = True,
        min_time: float = None,
        max_time: float = None,
        timestep: float = None,
        tod_preprocessing: dict = {},
        map_postprocessing: dict = {},
        k=0,
        progress_bars: bool = True,
    ):
        super().__init__(
            tods=tods,
            stokes=stokes,
            center=center,
            width=width,
            height=height,
            resolution=resolution,
            frame=frame,
            units=units,
            tod_preprocessing=tod_preprocessing,
            map_postprocessing=map_postprocessing,
            min_time=min_time,
            max_time=max_time,
            timestep=timestep,
            degrees=degrees,
            progress_bars=progress_bars,
        )

        # if center is None:
        #     center = np.degrees(get_center_phi_theta(*np.stack([tod.coords.center(frame="ra/dec") for tod in tods]).T))

        self.map_init_sigma = 0

        self.iteration = 0

        self.k = k

        self.binner = BinMapper(
            center=self.center,
            stokes=self.stokes,
            width=self.width,
            height=self.height,
            resolution=self.resolution,
            frame=self.frame,
            tods=self.tods,
            units=self.units,
            min_time=min_time,
            max_time=max_time,
            timestep=timestep,
            degrees=False,
            tod_preprocessing={
                "remove_modes": {"modes_to_remove": 1},
                "remove_spline": {"knot_spacing": 10, "remove_el_gradient": True},
            },
            map_postprocessing={},
            progress_bars=False,
        )

        # # self.update_deprojection()

        naive_sol = self.naive_map.ravel()

        init_sol = naive_sol
        # init_sol = 1e-5 * np.nanstd(naive_sol) * np.random.standard_normal(size=naive_sol.shape)

        self.mask = ~torch.tensor(init_sol).isnan()
        self.sol = MaximumLikelihoodSolution(data=torch.tensor(init_sol).float(), mask=self.mask)

        self.update_noise_model()

    def update_noise_model(self):
        self.tod_list = []

        pbar = tqdm(enumerate(self.tods), desc="Computing noise model", total=len(self.tods))

        for tod_index, tod in pbar:
            pbar.set_postfix(tod=f"{tod_index + 1}/{len(self.tods)}")

            t = {"tod": tod}

            t["k"] = max(self.k, 1)

            t["nd"], t["nt"] = tod.shape

            X = tod.dets.offsets
            rel_X = 2 * (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0)) - 1

            i, j = 0, 0
            mn_list = []
            while len(mn_list) < self.k:
                mn_list.append((i - j, j))
                j += 1
                if j > i:
                    i += 1
                    j = 0

            mode_list = []
            for m, n in mn_list:
                mode = sp.special.legendre(m)(rel_X[:, 0]) * sp.special.legendre(n)(rel_X[:, 1])
                mode_list.append(torch.tensor(mode, dtype=torch.complex64))

            t["a"] = torch.stack(mode_list, dim=1) if mode_list else torch.ones((t["nd"], 1), dtype=torch.complex64)

            ntip = int(0.5 * t["tod"].sample_rate)
            d = tod.signal.compute()
            d = remove_ends(d)
            # d = d - np.linspace(d[:, :ntip].mean(axis=-1), d[:, -ntip:].mean(axis=-1), d.shape[-1]).T

            # a, b = decompose(d, k=1)
            t["d"] = torch.tensor(d, dtype=torch.float)

            t["w"] = torch.tensor(sp.signal.windows.tukey(M=t["d"].shape[-1], alpha=0.1), dtype=torch.float)
            t["f"] = np.fft.fftfreq(n=t["d"].shape[-1], d=1 / t["tod"].sample_rate)
            t["abs_f"] = np.abs(t["f"])

            # def pointing_matrix(self, coords: Coordinates, dets: Array):
            #     values, pixel_index, sample_index = self.compute_pointing_matrix_sparse_indices(coords=coords, dets=dets)
            #     return sp.sparse.csr_array(
            #         (values, (sample_index, pixel_index)), shape=(coords.size, self.data.size)
            #     )

            # compute the pointing tensor (but the torch version)
            # sample_index, pixel_index, n_pixels = self.map.compute_pointing_matrix_sparse_indices(t["tod"].coords)
            values, pixel_index, sample_index = self.map.compute_pointing_matrix_sparse_indices(
                coords=t["tod"].coords, dets=t["tod"].dets
            )
            indices = torch.stack(
                [torch.tensor(sample_index, dtype=torch.long), torch.tensor(pixel_index, dtype=torch.long)], dim=0
            )
            t["P"] = torch.sparse_coo_tensor(
                indices=indices, values=torch.tensor(values, dtype=torch.float), size=(t["tod"].coords.size, self.map_size)
            )

            # project the fitted map out of the data
            t["d_no_map"] = t["d"] - (t["P"] @ self.sol.x()).reshape(t["d"].shape).detach()
            t["fd"] = torch.fft.fft(t["w"] * torch.tensor(remove_ends(t["d"].numpy())))
            # t["a"], t["b"] = [torch.tensor(x, dtype=torch.complex64) for x in decompose(t["fd"].numpy(), k=self.k)]
            # t["a"] = torch.ones((t["d"].shape[0], self.k), dtype=torch.complex64)

            # t["b"] = torch.linalg.inv(t["a"].T @ t["a"]) @ t["a"].T @ t["fd"]
            # if self.k > 0 else torch.zeros((1, t["nt"]), dtype=torch.complex64)
            # t["b"] = torch.tensor(sp.ndimage.gaussian_filter1d(t["b"], sigma=1, axis=-1))

            t["residual_ps"] = (t["fd"]).square().abs().float()

            n_bin = 32
            mid_f = np.geomspace(t["abs_f"][t["abs_f"] > 0].min() * 0.99, t["abs_f"].max() * 1.01, n_bin)
            dlogf = np.exp(np.gradient(np.log(mid_f)).mean())
            bin_f = np.geomspace(mid_f[0] / np.sqrt(dlogf), mid_f[-1] * np.sqrt(dlogf), n_bin + 1)
            bin_y = sp.stats.binned_statistic(t["abs_f"], t["residual_ps"].log(), bins=bin_f).statistic

            use = ~np.isnan(bin_y).any(axis=0)
            t["int_ps"] = torch.tensor(
                sp.interpolate.interp1d(
                    mid_f[use], np.median(bin_y[:, use], axis=0), axis=0, bounds_error=False, fill_value="extrapolate"
                )(t["abs_f"]),
                dtype=torch.float,
            ).exp()  # * torch.ones_like(t["d"])

            # t["V"] = (t["a"].T.unsqueeze(-1) * t["b"].unsqueeze(-2)).reshape(t["k"], -1)
            # t["ViA"] = (t["V"].reshape(t["k"], *t["d"].shape) / t["int_ps"]).reshape(t["k"], -1)
            # t["B"] = torch.linalg.inv(torch.eye(t["k"]) - t["ViA"] @ t["V"].T)

            t["PNd"] = t["P"].T @ self.apply_inverse_noise_covariance(t["d"], t)

            self.tod_list.append(t)

    def apply_inverse_noise_covariance(self, d, t):
        fd = torch.fft.fft(t["w"] * d, axis=-1)  # (n x t)

        Nfd1 = fd / t["int_ps"]

        if self.k > 0:
            Nfd2 = (t["ViA"].T @ (t["B"] @ (t["ViA"] @ fd.ravel()))).reshape(fd.shape)
        else:
            Nfd2 = 0.0

        return torch.fft.ifft(Nfd1 - Nfd2).real.ravel()

    @property
    def naive_map(self):
        if not hasattr(self, "_naive_map"):
            self._naive_map = self.binner.run()
        return self._naive_map.data.compute().ravel()

    def forward(self, t):
        return t["P"].T @ self.apply_inverse_noise_covariance((t["P"] @ self.sol.x()).reshape(t["d"].shape), t)

    def loss(self):
        return sum([(self.forward(t) - t["PNd"]).square().sum() for t in self.tod_list])

    @property
    def map(self):
        return ProjectionMap(
            data=self.sol.x().detach().numpy().reshape(self.map_shape),
            weight=np.ones(self.map_shape),
            stokes=self.stokes,
            t=self.t,
            nu=self.nu,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame.name,
            units=self.tod_units,
            beam=self.beam,
        ).to(self.map_units)

    def fit(self, epochs: int = 4, steps_per_epoch: int = 64, lr: float = 1e-1):
        if not hasattr(self, "optimizer"):
            self.optimizer = torch.optim.Adam(self.sol.parameters(), lr=lr)

        self.sol.train()

        for epoch in range(epochs):
            pbar = trange(steps_per_epoch, bar_format="{l_bar}{bar:16}{r_bar}{bar:-10b}", desc=f"epoch {epoch + 1}/{epochs}")

            try:
                for step in pbar:
                    loss = self.loss()
                    loss.backward()
                    self.optimizer.step()
                    self.optimizer.zero_grad()
                    pbar.set_postfix(loss=f"{loss.item():.03e}")

            except KeyboardInterrupt:
                break



from __future__ import annotations

import maria
import pytest


@pytest.mark.parametrize("nside", [256, 512, 1024, 2048])
def test_generate_cmb(nside):
    cmb = maria.cmb.generate_cmb(nside=nside)
    # cmb.plot()


def test_cmb_calibration():
    plan = maria.Plan.generate(
        scan_pattern="daisy",
        scan_options={"radius": 10, "speed": 1},  # in degrees
        duration=120,  # in seconds
        sample_rate=50,  # in Hz
        scan_center=(150, 50),
        jitter=0,
        frame="az/el",
    )

    sim = maria.Simulation(
        instrument="test/1deg",
        plans=plan,
        site="cerro_toco",
        cmb="generate",
        noise=False,
    )

    tod = sim.run()[0].to("uK_CMB")

    # should be around 110 uK
    cmb_rms_uK = tod.signal.std(axis=-1).compute()

    assert all(cmb_rms_uK > 80)
    assert all(cmb_rms_uK < 150)



from __future__ import annotations

import logging
import os
import time as ttime

import arrow
import numpy as np
import scipy as sp
from jax import scipy as jsp
from numpy.linalg import LinAlgError
from tqdm import tqdm

from ..functions import approximate_normalized_matern
from ..instrument import Instrument
from ..io import DEFAULT_BAR_FORMAT, humanize_time
from ..spectrum import AtmosphericSpectrum
from ..units import Quantity
from ..utils import compute_aligning_transform
from ..weather import Weather
from .extrusion import generate_layers
from .process import AutoregressiveProcess

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")

SUPPORTED_MODELS_LIST = ["2d", "3d"]

DEFAULT_ATMOSPHERE_KWARGS = {}


class Atmosphere:
    def __init__(
        self,
        model: str = "2d",
        timestamp: float = arrow.now().timestamp(),
        region: str = "princeton",
        altitude: float = None,
        weather_quantiles: dict = {},
        weather_kwargs: dict = {},
        weather_source: str = "era5",
        spectrum_source: str = "am",
        pwv_rms_frac: float = 0.03,
        max_height: float = 5e3,
        timestep: float = None,
        disable_progress_bars: bool = False,
    ):
        if model not in SUPPORTED_MODELS_LIST:
            raise ValueError(
                f"Invalid model '{model}'. Supported models are {SUPPORTED_MODELS_LIST}.",
            )

        self.spectrum = AtmosphericSpectrum(
            region=region,
            source=spectrum_source,
        )

        self.weather = Weather(
            time=timestamp,
            region=region,
            altitude=altitude,
            quantiles=weather_quantiles,
            override=weather_kwargs,
            source=weather_source,
        )

        self.pwv_rms_frac = pwv_rms_frac

        self.angular = False

        self.model = model
        self.max_height = max_height
        self.timestep = timestep

        self.disable_progress_bars = disable_progress_bars

        self._initialized = False

    def initialize(self, instrument, boresight, site):
        """
        Simulate a realization of PWV.
        """

        self.layers = generate_layers(
            instrument=instrument,
            boresight=boresight,
            weather=self.weather,
            site=site,
            mode=self.model,
            angular=self.angular,
            max_height=self.max_height,
        )

        if self.timestep is None:
            min_fwhm = instrument.dets.angular_fwhm(z=self.max_height).min()
            max_wind = Quantity((self.layers.wind_speed / self.layers.h).values, "rad/s").max()
            self.timestep = (min_fwhm / max_wind).s

        self.boresight = boresight.downsample(timestep=self.timestep)
        self.coords = self.boresight.broadcast(
            instrument.dets.offsets,
            frame="az/el",
        )

        # this is a smaller version of the sim coords
        outer_coords = self.boresight.broadcast(
            instrument.dets.outer().offsets,
            frame="az/el",
        )

        self.processes = {}

        center = outer_coords.center(frame="az/el")

        for process_index in tqdm(
            sorted(np.unique(self.layers.process_index)),
            desc="Constructing atmosphere",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
        ):
            process_init_s = ttime.monotonic()

            in_process = self.layers.process_index == process_index
            process_layers = self.layers.loc[in_process]

            # all processes must have the same velocity! so we compute it here.
            # vx and vy must have shape (n_time,)

            if process_layers.angular.any():
                vx = (
                    process_layers.wind_east.values[:, None] * np.cos(self.boresight.az)
                    - process_layers.wind_north.values[:, None] * np.sin(self.boresight.az)
                ) / process_layers.h.values[:, None]
                vy = (
                    -process_layers.wind_north.values[:, None] * np.cos(self.boresight.az)
                    - process_layers.wind_east.values[:, None] * np.sin(self.boresight.az)
                ) / (process_layers.h.values[:, None] * np.sin(self.boresight.el))
            else:
                vx = process_layers.wind_east.values[:, None] * np.ones(
                    self.boresight.shape[-1],
                )
                vy = process_layers.wind_north.values[:, None] * np.ones(
                    self.boresight.shape[-1],
                )

            w = (process_layers.absolute_humidity * process_layers.temperature * process_layers.divergence).values
            vx = (w[:, None] * vx).sum(axis=0) / w.sum()
            vy = (w[:, None] * vy).sum(axis=0) / w.sum()
            vz = np.zeros(vx.shape)

            if len(process_layers) > 1:
                process_layers = process_layers.loc[process_layers.index[[0, -1]]]

            # here we want enough points for a hull
            # we have made sure that any hull for this subpointing is a cover for the whole pointing

            process_points_for_hull_list = []

            for i, (layer_index, layer) in enumerate(process_layers.iterrows()):
                if layer.angular:
                    layer_x, layer_y = outer_coords.offsets(
                        center=center,
                        frame="az/el",
                    )

                else:
                    p = outer_coords.project(z=layer.h)
                    p += np.cumsum(self.timestep * np.c_[vx, vy, vz][None], axis=-2)

                    process_points_for_hull_list.append(p[None])

                    layer_x = layer.z * np.sin(outer_coords.az) * np.cos(outer_coords.el)
                    layer_y = layer.z * np.cos(outer_coords.az) * np.cos(outer_coords.el)

                layer_x += np.cumsum(self.timestep * vx)
                layer_y += np.cumsum(self.timestep * vy)

            process_points_for_hull = np.concatenate(
                process_points_for_hull_list,
                axis=0,
            ).reshape(-1, 3)
            process_points_for_hull[..., 2] += 1e-6 * np.random.standard_normal(
                process_points_for_hull[..., 2].shape,
            )
            # process_points_for_hull = process_points_for_hull.reshape(-1, 3)

            process_init_s = ttime.monotonic()

            transform = compute_aligning_transform(
                process_points_for_hull,
                signature=(True, True, False),
            )
            tp = process_points_for_hull @ transform
            triangulation = sp.spatial.Delaunay(tp[..., 1:])

            logger.debug(f"Computed process bounds in {humanize_time(ttime.monotonic() - process_init_s)}.")

            min_tx, min_ty, min_tz = tp.min(axis=0)
            max_tx, max_ty, max_tz = tp.max(axis=0)

            cross_section_points_list = []
            layer_labels = []

            for i, (layer_index, layer) in enumerate(
                self.layers.loc[in_process].iterrows(),
            ):
                # this algorithm probably works
                res = layer.res
                wide_lp_x_dense = np.linspace(min_ty, max_ty, 10000)
                wide_lp_z_dense = layer.h * np.ones(len(wide_lp_x_dense))
                wide_lp_dense = np.c_[wide_lp_x_dense, wide_lp_z_dense]
                interior = triangulation.find_simplex(wide_lp_dense) > -1
                # interior = interior if interior.any() else np.ones_like(interior, dtype=bool)
                lp_dense_x = wide_lp_x_dense[interior]
                lp_x_min = lp_dense_x.min() - 2 * res
                lp_x_max = lp_dense_x.max() + 2 * res
                n_lp = np.maximum(3, int((lp_x_max - lp_x_min) / res))
                lp_x = np.linspace(lp_x_min, lp_x_max, n_lp)
                lp_z = layer.h * np.ones(len(lp_x))
                lp = np.c_[lp_x, lp_z]
                n_lp = len(lp)

                layer_labels.extend(n_lp * [layer_index])
                cross_section_points_list.append(lp)

            cross_section_points = np.concatenate(cross_section_points_list, axis=0)

            extrusion_points = np.arange(
                min_tx - 2 * self.layers.res.min(),
                max_tx + 2 * self.layers.res.min(),
                self.layers.res.min(),
            )

            outer_scale = np.maximum(1e3, 300 + process_layers.h.mean() / 10)

            matern_kwargs = {"nu": 1 / 3, "r0": outer_scale} if self.model == "3d" else {"nu": 5 / 6, "r0": outer_scale}

            process = AutoregressiveProcess(
                cross_section=cross_section_points,
                extrusion=extrusion_points,
                callback=approximate_normalized_matern,
                callback_kwargs=matern_kwargs,
                jitter=1e-8,
            )

            process.layers = self.layers.loc[in_process]
            process.labels = np.array(layer_labels)
            process.transform = transform
            process.extrusion_res = np.gradient(extrusion_points).mean()
            process.vx = vx
            process.vy = vy
            process.tp = tp

            success = False
            for jitter in [1e-8, 1e-6, 1e-4]:
                try:
                    process.compute_covariance_matrices()
                    success = True
                    break
                except LinAlgError as e:
                    logger.debug(f"Singular covariance matrix with jitter={jitter}")

            if not success:
                raise LinAlgError("Covariance matrix was singular even with max jitter.")

            self.processes[int(process_index)] = process

        self._initialized = True

    def __repr__(self):
        return f"""Atmosphere({len(self.processes)} processes with {len(self.layers)} layers):
├ spectrum:
│   region: {self.spectrum.region}
└ weather:
    region: {self.weather.region}
    altitude: {Quantity(self.weather.base_altitude, "m")}
    time: {self.weather.local_time.format("MMM D HH:mm:ss ZZ")}
    pwv[mean, rms]: ({Quantity(self.weather.pwv, "mm")}, {Quantity(self.pwv_rms_frac * self.weather.pwv, "mm")})"""

    def simulate_pwv(self, instrument: Instrument = None):
        if not self._initialized:
            raise RuntimeError("Atmosphere must be initialized with a simulation.")

        pp = self.coords.project(z=1)

        for k, process in tqdm(
            self.processes.items(),
            desc="Generating turbulence",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
        ):
            process_s = ttime.monotonic()
            process.run()
            logger.debug(f"Ran atmospheric process {k} in {humanize_time(ttime.monotonic() - process_s)}")

        self.zenith_scaled_pwv = self.weather.pwv * np.ones(shape=pp.shape[:-1])

        with tqdm(
            total=len(self.layers),
            desc="Sampling turbulence",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
        ) as pbar:
            for k, process in self.processes.items():
                wind_vector = np.c_[process.vx, process.vy, np.zeros(process.vx.shape)]
                translation = np.cumsum(self.timestep * wind_vector[None], axis=-2)

                for i, (layer_index, layer) in enumerate(
                    process.layers.iterrows(),
                ):
                    layer_s = ttime.monotonic()

                    layer_mask = process.labels == layer_index

                    beam_fwhm = (
                        instrument.dets.physical_fwhm(
                            layer.z,
                        )
                        .mean()
                        .meters
                        if instrument
                        else 0
                    )
                    beam_sigma = beam_fwhm / 2.355
                    extrusion_sigma = beam_sigma / process.extrusion_res
                    x_sigma = beam_sigma / layer.res

                    smoothed_layer_pwv = sp.ndimage.gaussian_filter(
                        process.values[:, layer_mask],
                        sigma=(extrusion_sigma, x_sigma),
                    )

                    p = layer.h * pp + translation
                    transformed_lpp = p @ process.transform

                    # NOT JAX because linear interpolations leads to weird jagged things in Fourier space
                    # y = sp.interpolate.RegularGridInterpolator(
                    #     (
                    #         process.extrusion,
                    #         process.cross_section[layer_mask, 0],
                    #     ),
                    #     smoothed_layer_pwv,
                    #     method="cubic",
                    # )(transformed_lpp[..., :2])

                    y = jsp.interpolate.RegularGridInterpolator(
                        (
                            process.extrusion,
                            process.cross_section[layer_mask, 0],
                        ),
                        smoothed_layer_pwv,
                        method="linear",
                    )(transformed_lpp[..., :2])

                    if np.isnan(y).any():
                        raise RuntimeError(f"Layer {layer_index} in process {k} introduced nans into PWV simulation.")

                    # y = y

                    self.zenith_scaled_pwv += layer.pwv_rms * y

                    pbar.update(1)

                    logger.debug(
                        f"Sampled layer {layer_index} (h={Quantity(layer.h, 'm')}) with shape "
                        f"{smoothed_layer_pwv.shape} in {humanize_time(ttime.monotonic() - layer_s)}"
                    )



from __future__ import annotations

import copy
import glob
import logging
import os
import uuid
from typing import Mapping

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from matplotlib.collections import EllipseCollection
from matplotlib.patches import Patch
from scipy.spatial import QhullError

from ..band import Band, BandList  # noqa
from ..beam import compute_angular_fwhm
from ..units import Quantity
from ..utils import HEX_CODE_LIST, compute_diameter, flatten_config, get_rotation_matrix_2d, read_yaml
from .generation import generate_2d_pattern

here, this_filename = os.path.split(__file__)


PER_DET_KWARGS = [
    "sky_x",
    "sky_y",
    "baseline_x",
    "baseline_y",
    "baseline_z",
    "pol_angle",
    "band_name",
    "bath_temp",
]


ALLOWED_ARRAY_KWARGS = [
    "band",
    "bands",
    "max_baseline",
    "baseline_offset",
    "baseline_spacing",
    "bath_temp",
    "beam_spacing",
    "degrees",
    "field_of_view",
    "file",
    "focal_plane_offset",
    "key",
    "n",
    "n_col",
    "n_row",
    "name",
    "packing",
    "polarized",
    "primary_size",
    "rotation",
    "shape",
    *PER_DET_KWARGS,
]


DET_COLUMN_TYPES = {
    "array_name": str,
    "uid": str,
    "base_det_index": int,
    "band_name": str,
    "band_center": float,
    "sky_x": float,
    "sky_y": float,
    "baseline_x": float,
    "baseline_y": float,
    "baseline_z": float,
    "pol_angle": float,
    "pol_label": str,
    "primary_size": float,
    "bath_temp": float,
    "time_constant": float,
    "white_noise": float,
    "pink_noise": float,
    "efficiency": float,
}


ARRAY_CONFIGS = {}
for path in glob.glob(f"{here}/configs/*.yml"):
    key = os.path.split(path)[1].split(".")[0]
    ARRAY_CONFIGS[key] = read_yaml(path)
ARRAY_CONFIGS = flatten_config(ARRAY_CONFIGS)

all_arrays = list(ARRAY_CONFIGS.keys())


def get_array_config(key=None, **kwargs):
    c = {}
    if key:
        if key not in ARRAY_CONFIGS:
            raise KeyError(f"'{key}' is not a valid array name.")
        c = {"name": key, **ARRAY_CONFIGS[key]}
    c.update(kwargs)
    return c


def get_array(key):
    return Array.from_config(get_array_config(key=key))


class Array:
    def __init__(self, name: str, dets: pd.DataFrame, bands: BandList, config: dict = {}):
        self.name = name  # or str(uuid.uuid4())
        self.dets = dets
        self.dets.index = np.arange(len(self.dets.index))
        self.dets.loc[:, "array_name"] = self.name

        self.bands = BandList([band for band in bands if band.name in self.dets.band_name.values])
        self.config = config

        # for band_attr, det_attr in {
        #     "center": "band_center",
        #     "width": "band_width",
        # }.items():

        # self.dets.loc[:, "band_center"] = getattr(self, band_attr.Hz)

    def split(self):
        array_list = []
        for array_name in sorted(np.unique(self.dets.array_name.values)):
            array_dets = self.dets.loc[self.dets.array_name.values == array_name]
            array_bands = [band for band in self.bands if band.name in array_dets.band_name.values]
            array_list.append(Array(name=array_name, dets=array_dets, bands=array_bands))
        return ArrayList(array_list)

    def mask(self, **kwargs):
        mask = np.ones(len(self.dets)).astype(bool)
        for k, v in kwargs.items():
            mask &= self.dets.loc[:, k].values == v
        return mask

    def subset(self, **kwargs):
        return self._subset(self.mask(**kwargs))

    def _subset(self, mask):
        dets = self.dets.loc[mask]
        return Array(name=self.name, dets=dets, bands=[b for b in self.bands if b.name in dets.band_name.values])  # noqa

    def one_detector_from_each_band(self):
        first_det_mask = np.isin(
            np.arange(self.n),
            np.unique(self.band_name, return_index=True)[1],
        )
        return self._subset(mask=first_det_mask)

    def outer(self):
        try:
            hull = sp.spatial.ConvexHull(self.offsets)
        except QhullError:
            return self
        outer_dets_mask = np.isin(np.arange(self.n), hull.vertices)
        return self._subset(mask=outer_dets_mask)

    @property
    def n(self):
        return len(self.dets)

    @property
    def offsets(self):
        return np.c_[self.sky_x, self.sky_y]

    @property
    def baselines(self):
        return np.c_[self.baseline_x, self.baseline_y, self.baseline_z]

    @property
    def field_of_view(self):
        return Quantity(compute_diameter(self.offsets), units="rad")

    @property
    def max_baseline(self):
        return Quantity(compute_diameter(self.baselines), units="m")

    @property
    def index(self):
        return self.dets.index.values

    @property
    def ubands(self):
        return list(self.bands.keys())

    @property
    def fwhm(self):
        """
        Returns the angular FWHM (in radians) at infinite distance.
        """
        return self.angular_fwhm(z=np.inf)

    @property
    def beams(self):
        fwhm = self.fwhm.radians
        return np.stack([fwhm, fwhm, np.zeros_like(fwhm)], axis=1)

    def mueller(self):
        """
        For transforming Stoke's parameters. $theta$ = 0 is the horizontal
        """
        a = self.pol_angle
        m = np.stack(
            [
                np.where(np.isnan(a), np.sqrt(2), 1),
                np.where(np.isnan(a), 0, np.cos(2 * a)),
                np.where(np.isnan(a), 0, np.sin(2 * a)),
                np.zeros_like(a),
            ],
            axis=1,
        )
        return 0.5 * m[..., None] * m[..., None, :]

    def stokes_weight(self):
        return self.mueller()[:, 0]

    def angular_fwhm(self, z=np.inf):  # noqa F401
        """
        Angular beam width (in radians) as a function of depth (in meters)
        """
        return Quantity(compute_angular_fwhm(z=z, fwhm_0=self.primary_size, n=1, nu=self.band_center), "rad")

    def physical_fwhm(self, z):
        """
        Physical beam width (in meters) as a function of depth (in meters)
        """
        return Quantity(z * self.angular_fwhm(z).rad, "m")

    @property
    def band_center(self):
        values = np.zeros(shape=self.n, dtype=float)
        for band in self.bands:
            values[self.band_name == band.name] = band.center.Hz
        return values

    @property
    def band_width(self):
        values = np.zeros(shape=self.n, dtype=float)
        for band in self.bands:
            values[self.band_name == band.name] = band.width.Hz
        return values

    def passband(self, nu):
        _nu = np.atleast_1d(nu)
        PB = np.zeros((len(self.dets), len(_nu)))
        for band in self.bands:
            PB[self.band_name == band.name] = band.passband(_nu)
        return PB

    def __call__(self, band: str = None):
        mask = True
        if band:
            mask &= self.band_name == band
        return Array(name=self.name, dets=self.dets.loc[mask], bands=self.bands)

    def __getattr__(self, attr):
        if attr in self.dets.columns:
            return self.dets.loc[:, attr].values.astype(DET_COLUMN_TYPES[attr])

        if all(hasattr(band, attr) for band in self.bands):
            values = np.zeros(shape=self.n, dtype=float)
            for band in self.bands:
                values[self.band_name == band.name] = getattr(band, attr)
            return values

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __len__(self):
        return len(self.dets)

    @property
    def polarized(self):
        return not np.isnan(self.pol_angle).all()

    def filling(self):
        return {
            "n": self.n,
            "FOV": self.field_of_view,
            "baseline": self.max_baseline,
            "bands": f"[{','.join(self.bands.name)}]",
            "polarized": self.polarized,
        }

    def summary(self):
        series = pd.Series(self.filling())
        series.name = self.name
        return series

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join([f'{k}={v}' for k, v in self.filling().items()])})"

    def __getitem__(self, key):
        return Array(name=self.name, dets=self.dets[key], bands=self.bands)

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls.from_config(kwargs)

    @classmethod
    def from_config(cls, config):
        c = copy.deepcopy(config)

        degrees = c.get("degrees", True)

        if "array_name" in config:
            c.update(get_array_config(key=c.pop("array_name")))

        logging.debug(f"Generating array with config {c}")

        disallowed_array_kwargs = [key for key in c if key not in ALLOWED_ARRAY_KWARGS]
        if disallowed_array_kwargs:
            raise ValueError(f"Invalid array kwargs {disallowed_array_kwargs}.")

        if "file" in c:
            file = c.pop("file")
            try:
                dets = pd.read_csv(f"{here}/{file}")
            except FileNotFoundError:
                dets = pd.read_csv(file)

            for col in dets.columns:
                c[col] = dets[col].values

        bands = None
        if "bands" in c:
            bands = BandList(c["bands"])
        elif "band" in c:
            bands = BandList([Band(c["band"])])
        elif "band_name" not in c:
            raise ValueError("Missing parameter 'bands'.")

        # check that these match known bands
        if "band_name" in c:
            if bands is not None:
                for band_name in np.unique(c["band_name"]):
                    if band_name not in bands.name:
                        raise ValueError(
                            f"Band name {band_name} in the supplied 'band_name' parameter does "
                            f"not match any band in the supplied 'bands' parameter ()."
                        )
            else:
                bands = BandList(sorted(list(np.unique(c["band_name"]))))

        if "primary_size" in c:
            primary_sizes = np.atleast_1d(c.pop("primary_size"))
        else:
            raise ValueError("Missing array parameter 'primary_size'.")

        if degrees:
            for param in ["sky_x", "sky_y", "field_of_view", "pol_angle", "rotation"]:
                if param in c:
                    c[param] = np.radians(c[param])

        if ("sky_x" in c) and ("sky_y" in c):
            if (c["sky_x"].ndim != 1) or (c["sky_y"].ndim != 1):
                raise ValueError("Parameters 'sky_{x,y}' must be broadcastable one-dimensional arrays.")
        elif ("baseline_x" in c) and ("baseline_y" in c):
            if (c["baseline_x"].ndim != 1) or (c["baseline_y"].ndim != 1):
                raise ValueError("Parameters 'baseline_{x,y}' must be broadcastable one-dimensional arrays.")

        else:
            max_resolution = max(
                [
                    compute_angular_fwhm(primary_size, z=np.inf, nu=band.center.Hz)
                    for band in bands
                    for primary_size in primary_sizes
                ]
            )
            # we need:
            # - two of (n-like, field_of_view, beam_spacing), or
            # - two of (n-like, max_baseline, baseline_spacing)

            pattern_kwargs = {}

            n_kwargs = {k: c.get(k) for k in ["n", "n_col", "n_row"] if c.get(k) is not None}
            n_explicit = ("n" in n_kwargs) or (("n_col" in n_kwargs) and ("n_row" in n_kwargs))
            n_focal_plane_kwargs = sum([n_explicit, "field_of_view" in c, "focal_plane_spacing" in c])
            n_baseline_kwargs = sum(
                [
                    n_explicit,
                    "max_baseline" in c,
                    "baseline_spacing" in c,
                ]
            )

            # this is not supposed to be elegant, it is supposed to be easy to understand
            if n_explicit:
                pattern_kwargs.update(n_kwargs)

                # handle the case of a single detector
                if (n_kwargs.get("n") == 1) or ((n_kwargs.get("n_row") == 1) and (n_kwargs.get("n_col") == 1)):
                    mode = "focal_plane"
                    pattern_kwargs["spacing"] = 0.0

                elif ("field_of_view" in c) or ("beam_spacing" in c):
                    # we can only use one if n is explicit
                    mode = "focal_plane"
                    if "field_of_view" in c:
                        pattern_kwargs["max_diameter"] = c["field_of_view"]
                    else:
                        pattern_kwargs["spacing"] = c.get("beam_spacing", 1.5) * max_resolution

                elif "max_baseline" in c:
                    mode = "baseline"
                    pattern_kwargs["max_diameter"] = c["max_baseline"]

                elif "baseline_spacing" in c:
                    mode = "baseline"
                    pattern_kwargs["spacing"] = c["baseline_spacing"]

                else:
                    raise ValueError(
                        "Invalid array spec: if detector counts are explicit, you must supply either 'field_of_view' or "
                        "either 'max_baseline' or 'baseline_spacing' for an interferometer array."
                    )

            else:
                pattern_kwargs = {}
                if "field_of_view" in c:
                    mode = "focal_plane"
                    pattern_kwargs["max_diameter"] = c["field_of_view"]
                    pattern_kwargs["spacing"] = c.get("beam_spacing", 1.5) * max_resolution

                elif ("max_baseline" in c) and ("baseline_spacing" in c):
                    mode = "baseline"
                    pattern_kwargs["max_diameter"] = c["max_baseline"]
                    pattern_kwargs["spacing"] = c["baseline_spacing"]

                else:
                    raise ValueError(
                        "Invalid array spec: if detector counts are implicit, you must supply either 'field_of_view' or "
                        "both 'max_baseline' or 'baseline_spacing' for an interferometer array."
                    )

            X = generate_2d_pattern(
                **pattern_kwargs,
                shape=c.get("shape", "hexagon"),
                packing=c.get("packing", "triangular"),
                rotation=c.get("rotation", 0),
            )

            if mode == "focal_plane":
                c["sky_x"], c["sky_y"] = X[..., 0], X[..., 1]
            else:
                c["baseline_x"], c["baseline_y"] = X[..., 0], X[..., 1]

        dets = pd.DataFrame({col: c[col] for col in c if col in PER_DET_KWARGS})
        dets.loc[:, "base_det_index"] = np.arange(len(dets))
        dets.loc[:, "primary_size"] = primary_sizes * np.ones(len(dets))

        for key, default in [("bath_temp", 0), ("time_constant", 0)]:
            if (key not in dets) or (key in c):
                dets[key] = np.array(c.get(key, default) * np.ones(len(dets)))

        baseline_offset = c.get("baseline_offset", (0.0, 0.0, 0.0))
        focal_plane_offset = c.get("focal_plane_offset", (0.0, 0.0))

        for i in range(3):
            dim = "xyz"[i]
            if f"baseline_{dim}" not in dets.columns:
                dets.loc[:, f"baseline_{dim}"] = 0.0
            dets.loc[:, f"baseline_{dim}"] += baseline_offset[i]
            if dim == "z":
                continue
            if f"sky_{dim}" not in dets.columns:
                dets.loc[:, f"sky_{dim}"] = 0.0
            dets.loc[:, f"sky_{dim}"] += np.radians(focal_plane_offset[i])

        if "pol_angle" not in dets:
            if c.get("polarized", False):
                dets.loc[:, "pol_angle"] = np.random.uniform(
                    low=0,
                    high=np.pi,
                    size=len(dets),
                )
                dets.loc[:, "pol_label"] = "A"
                other_dets = dets.copy()
                other_dets.loc[:, "pol_angle"] = (dets.pol_angle + np.pi / 2) % np.pi
                other_dets.loc[:, "pol_label"] = "B"
                dets = pd.concat([dets, other_dets])

            else:
                dets.loc[:, "pol_angle"] = np.nan

        if "band_name" not in dets.columns:
            band_dets_list = []
            for band in bands:
                band_dets = dets.copy()
                band_dets.loc[:, "band_name"] = band.name
                band_dets_list.append(band_dets)
            dets = pd.concat(band_dets_list)

        dets = dets.sort_values(["band_name", "base_det_index"], ascending=True)

        for col in dets.columns:
            dets[col] = dets[col].astype(DET_COLUMN_TYPES.get(col, str))

        return cls(dets=dets, bands=bands, name=c.get("name", str(uuid.uuid4())[:8]))

    def plot(self, z=np.inf, plot_baseline="infer", plot_pol_angles=False):
        # if plot_baseline == "infer":
        #     plot_baseline = self.dets.max_baseline > 0

        # if plot_baseline:
        fig, (focal_ax, band_ax) = plt.subplots(1, 2, figsize=(10, 5), dpi=256, constrained_layout=True)

        i = 0
        legend_handles = []
        band_legend_handles = []

        focal_plane_units = Quantity(self.offsets, "rad").units
        resolution_units = Quantity(self.fwhm, "rad").units

        nu_min, nu_max = self.bands.nu_min, self.bands.nu_max

        nu = Quantity(np.linspace(self.bands.nu_min.Hz, self.bands.nu_max.Hz, 1024), "Hz")
        nu_min, nu_max = nu.value.min(), nu.value.max()

        for ia, array in enumerate(self.split()):
            for ib, band in enumerate(array.bands):
                c = HEX_CODE_LIST[i % len(HEX_CODE_LIST)]

                band_array = array(band=band.name)

                fwhms = Quantity(band_array.angular_fwhm(z=z), "rad")
                offsets = Quantity(band_array.offsets, "rad")
                pol_angles = Quantity(band_array.pol_angle, "rad")
                # baselines = Quantity(band_array.baselines, "m")

                if plot_pol_angles and self.polarized:
                    dx = np.c_[-np.ones(band_array.n), np.ones(band_array.n)] / 2
                    dy = np.zeros((band_array.n, 2))

                    R = get_rotation_matrix_2d(pol_angles.rad)
                    dl = np.moveaxis(R @ np.stack([dx, dy], axis=1), 0, 2)
                    P = Quantity(offsets.rad.T[:, None] + fwhms.rad * dl, units="rad")
                    focal_ax.plot(*getattr(P, focal_plane_units), c=c, lw=2e0)

                focal_ax.add_collection(
                    EllipseCollection(
                        widths=getattr(fwhms, focal_plane_units),
                        heights=getattr(fwhms, focal_plane_units),
                        angles=0,
                        units="xy",
                        facecolors=c,
                        edgecolors="k",
                        lw=1e-1,
                        alpha=0.5,
                        offsets=getattr(offsets, focal_plane_units),
                        transOffset=focal_ax.transData,
                    ),
                )

                band_ax.plot(nu.value, band.passband(nu.Hz), color=c)

                legend_handles.append(
                    Patch(
                        label=f"{band.name} (n={band_array.n}, "
                        f"res={getattr(fwhms, resolution_units).mean():.01f} {resolution_units})",
                        color=c,
                    ),
                )
                band_legend_handles.append(Patch(label=rf"{band.name} ($\eta$={band.efficiency})", color=c))  # noqa

                focal_ax.scatter(
                    *getattr(offsets, focal_plane_units).T,
                    s=0,
                    color=c,
                )

                i += 1

        focal_ax.set_xlabel(rf"$\theta_x$ offset [{focal_plane_units}]")
        focal_ax.set_ylabel(rf"$\theta_y$ offset [{focal_plane_units}]")
        focal_ax.legend(handles=legend_handles, fontsize=8)

        band_ax.set_xlabel(rf"$\nu$ [${nu.u['math_name']}$]")
        band_ax.set_ylabel(rf"Passband response")
        band_ax.legend(handles=band_legend_handles, fontsize=8)

        band_ax.plot([nu_min, nu_max], [0, 0], c="k", lw=0.5, ls=":")
        band_ax.set_xlim(nu_min, nu_max)

        band_ax.yaxis.tick_right()
        band_ax.yaxis.set_label_position("right")

        xls, yls = focal_ax.get_xlim(), focal_ax.get_ylim()
        cen_x, cen_y = np.mean(xls), np.mean(yls)
        wid_x, wid_y = np.ptp(xls), np.ptp(yls)
        radius = 0.5 * np.maximum(wid_x, wid_y)

        margin = getattr(fwhms, focal_plane_units).max()

        focal_ax.set_xlim(cen_x - radius - margin, cen_x + radius + margin)
        focal_ax.set_ylim(cen_y - radius - margin, cen_y + radius + margin)


class ArrayList:
    def __init__(self, arrays: list[Mapping]):
        if isinstance(arrays, Mapping):
            # convert to a list
            arrays = [{"name": k, **v} for k, v in arrays.items()]

        if isinstance(arrays, list):
            unnamed_array_index = 1
            self.arrays = []
            for array in arrays:
                if isinstance(array, str):
                    self.arrays.append(get_array(array))
                elif isinstance(array, Mapping):
                    if "name" not in array:
                        array["name"] = f"array{unnamed_array_index}"
                        unnamed_array_index += 1
                    array_config = get_array_config(**array)
                    self.arrays.append(Array.from_config(array_config))
                elif isinstance(array, Array):
                    self.arrays.append(array)
                else:
                    raise ValueError("Arrays must be either a string or a mapping")
        elif isinstance(arrays, ArrayList):
            self.arrays = arrays.arrays
        else:
            raise ValueError("Each element of 'arrays' must be either an Array, a string, or a mapping.")

    def combine(self):
        array_dets = []
        for array in self.arrays:
            dets = copy.deepcopy(array.dets)
            dets.loc[:, "array_name"] = array.name
            array_dets.append(dets)
        return Array(name="", dets=pd.concat(array_dets), bands=self.bands)

    def one_detector_from_each_band(self):
        return ArrayList(arrays=[array.one_detector_from_each_band() for array in self.arrays])

    def outer(self):
        return ArrayList(arrays=[array.outer() for array in self.arrays])

    @property
    def field_of_view(self):
        return Quantity(compute_diameter(self.offsets), units="rad")

    @property
    def max_baseline(self):
        return Quantity(compute_diameter(self.baselines), units="m")

    @property
    def n(self):
        return sum([array.n for array in self.arrays])

    @property
    def dets(self):
        return pd.concat([array.dets for array in self.arrays])

    @property
    def bands(self):
        bands = []
        for array in self.arrays:
            for band in array.bands:
                if band not in bands:
                    bands.append(band)
        return BandList(bands)

    def angular_fwhm(self, z):  # noqa F401
        """
        Angular beam width (in radians) as a function of depth (in meters)
        """
        return Quantity(compute_angular_fwhm(z=z, fwhm_0=self.primary_size, n=1, nu=self.band_center), "rad")

    def physical_fwhm(self, z):
        """
        Physical beam width (in meters) as a function of depth (in meters)
        """
        return Quantity(z * self.angular_fwhm(z).rad, "m")

    def mask(self, **kwargs):
        return np.concatenate([array.mask(**kwargs) for array in self.arrays], axis=0)

    def subset(self, **kwargs):
        return ArrayList([array.subset(**kwargs).dets for array in self.arrays])

    def summary(self):
        return pd.concat([array.summary() for array in self.arrays], axis=1).T

    @property
    def array_name(self):
        return np.concatenate([array.n * [array.name] for array in self.arrays], axis=0)

    @property
    def offsets(self):
        return np.concatenate([array.offsets for array in self.arrays], axis=0)

    @property
    def baselines(self):
        return np.concatenate([array.baselines for array in self.arrays], axis=0)

    def passband(self, nu):
        return np.concatenate([array.passband(nu) for array in self.arrays], axis=0)

    def __getitem__(self, key):
        return self.arrays[key]

    def __getattr__(self, attr):
        try:
            return np.concatenate([getattr(array, attr) for array in self.arrays], axis=0)
        except Exception:
            pass
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __repr__(self):
        return self.summary().__repr__()

    def _repr_html_(self):
        return self.summary()._repr_html_()

    def __iter__(self):  # it has to be called this
        return iter(self.arrays)  # return the list's iterator

    def __len__(self):
        return len(self.arrays)



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from maria import Simulation
from maria.coords import Coordinates, unjitted_offsets_to_phi_theta
from maria.instrument import get_instrument
from maria.noise import generate_noise_with_knee
from maria.plan import get_plan
from maria.tod.tod import TOD

plt.close("all")


def test_tod_functions():
    n = 256

    time = 1.75e9 + np.arange(0, 600, 0.1)
    azim = np.radians(45) * np.ones(len(time))
    elev = np.radians(45) * np.ones(len(time))

    offsets = np.radians(1e1 * np.random.standard_normal(size=(n, 1, 2)))

    PT = unjitted_offsets_to_phi_theta(offsets, azim, elev)

    coords = Coordinates(phi=PT[..., 0], theta=PT[..., 1], t=time, frame="az/el")

    noise = generate_noise_with_knee(shape=coords.shape, sample_rate=1 / coords.timestep, knee=0.5)

    tod = TOD(data=dict(noise=noise), coords=coords)

    print(f"{tod.sample_rate =}")
    print(f"{tod.duration =}")
    print(f"{tod.boresight =}")


def test_tod_calibration_with_atmosphere():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        atmosphere="2d",
        noise=True,
    )

    tod = sim.run()[0]

    tod = tod.to("pW")
    tod = tod.to("K_RJ")


def test_tod_preprocessing_with_config():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        noise=True,
    )

    tod = sim.run()[0][::2, :10000]

    pp_config = {
        "window": {"name": "tukey", "kwargs": {"alpha": 0.25}},
        "filter": {"f_lower": 0.5},
        "remove_modes": {"modes_to_remove": 1},
        "remove_spline": {"knot_spacing": 0.5},
    }

    tod.process(config=pp_config)

    tod.twinkle(
        rate=2,
        max_frames=10,
        filename="/tmp/test_twinkle.gif",
    )


def test_tod_preprocessing_with_kwargs():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        noise=True,
    )

    tod = sim.run()[0]

    tod.process(window="tukey")


def test_tod_preprocessing_errors():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        noise=True,
    )

    tod = sim.run()[0]

    try:
        tod.process(f_lower="a")
        assert False
    except TypeError:
        pass


def test_tod_write_and_load():
    plan = get_plan(
        scan_pattern="daisy",  # scanning pattern
        scan_options={"radius": 2 / 60, "speed": 0.5 / 60},  # in degrees
        duration=600,  # integration time in seconds
        sample_rate=50,  # in Hz
        scan_center=(202.27211, 47.195277),  # position in the sky
        frame="az/el",
    )

    sim = Simulation(get_instrument("MUSTANG-2"), plans=plan, site="green_bank")
    tod = sim.run()[0]
    tod.to_fits("/tmp/sim_tod.fits")
    tod_loaded = TOD.from_fits("/tmp/sim_tod.fits", format="MUSTANG-2")



from __future__ import annotations

import logging
import time as ttime

import dask.array as da
import numpy as np
from tqdm import tqdm

from ..functions import approximate_normalized_matern
from ..io import humanize_time
from ..utils import fast_psd_inverse

logger = logging.getLogger("maria")

COV_MAT_JITTER = 1e-6


class AutoregressiveProcess:
    def __init__(
        self,
        cross_section,
        extrusion,
        callback=approximate_normalized_matern,
        callback_kwargs: dict = {},
        lookback_decay_rate: float = 2,
        jitter: float = 1e-6,
        MIN_SAMPLES_PER_LAYER: int = 4,
    ):
        self.cross_section = cross_section
        self.extrusion = extrusion
        self.callback = callback
        self.callback_kwargs = callback_kwargs
        self.lookback_decay_rate = lookback_decay_rate
        self.jitter = jitter
        self.MIN_SAMPLES_PER_LAYER = MIN_SAMPLES_PER_LAYER

        self.n_cross_section = len(self.cross_section)
        self.n_extrusion = len(self.extrusion)
        I, J = np.meshgrid(np.arange(self.n_cross_section), np.arange(self.n_extrusion))  # noqa

        points = np.c_[self.extrusion[J][..., None], self.cross_section[I]]

        extrusion_indices = [
            0,
            *(2 ** np.arange(0, np.log(self.n_extrusion) / np.log(2))).astype(int),
            self.n_extrusion - 1,
        ]

        self.extrusion_sample_index = []
        self.cross_section_sample_index = []
        for i, extrusion_index in enumerate(extrusion_indices):
            n_ribbon_samples = np.minimum(
                np.maximum(
                    int(self.n_cross_section * (2) ** -(i)),
                    self.MIN_SAMPLES_PER_LAYER,
                ),
                self.n_cross_section,
            )
            # cross_section_indices = sorted(np.random.choice(a=self.n_cross_section, size=n_ribbon_samples, replace=False))
            cross_section_indices = np.unique(
                np.linspace(0, self.n_cross_section - 1, n_ribbon_samples).astype(int),
            )
            self.cross_section_sample_index.extend(cross_section_indices)
            self.extrusion_sample_index.extend(
                np.repeat(extrusion_index, len(cross_section_indices)),
            )

        self.cross_section_sample_index = np.array(self.cross_section_sample_index)
        self.extrusion_sample_index = np.array(self.extrusion_sample_index)

        E_sample = points[
            self.extrusion_sample_index,
            self.cross_section_sample_index,
            0,
        ]
        X_sample = points[
            self.extrusion_sample_index,
            self.cross_section_sample_index,
            1,
        ]
        Y_sample = points[
            self.extrusion_sample_index,
            self.cross_section_sample_index,
            2,
        ]

        self.sample_points = np.c_[E_sample, X_sample, Y_sample]
        self.n_sample = len(self.sample_points)

        self.E_live_edge = points[0, :, 0] - np.gradient(self.extrusion).mean()
        self.X_live_edge = points[0, :, 1]
        self.Y_live_edge = points[0, :, 2]

        self.live_edge_points = np.c_[
            self.E_live_edge,
            self.X_live_edge,
            self.Y_live_edge,
        ]
        self.n_live_edge = len(self.live_edge_points)

        n_side_warn = 4000

        if self.n_sample > n_side_warn:
            logger.warning(
                f"A large covariance matrix (n_side={self.n_sample}) will be generated; inverting these matrices is very expensive.",  # noqa
            )

        self.points = da.asarray(points)

    def compute_covariance_matrices(self):
        # edge-edge upper {i,j}
        covariance_s = ttime.monotonic()
        i, j = np.triu_indices(self.n_live_edge, k=1)
        COV_E_E = np.eye(self.n_live_edge) + self.jitter
        COV_E_E[i, j] = self.callback(
            np.sqrt(
                np.square(self.live_edge_points[j] - self.live_edge_points[i]).sum(
                    axis=1,
                ),
            ),
            **self.callback_kwargs,
        )
        COV_E_E[j, i] = COV_E_E[i, j]
        COV_E_E += np.diag(COV_MAT_JITTER * np.diag(COV_E_E))  # add some jitter
        # logger.debug(
        #     f"Computed edge-edge covariance {COV_E_E.shape} in {1e3 * (ttime.monotonic() - start_time):.0f} ms.",
        # )

        # this one is explicit
        COV_E_S = self.callback(
            np.sqrt(
                np.square(
                    self.sample_points[None] - self.live_edge_points[:, None],
                ).sum(axis=2),
            ),
            **self.callback_kwargs,
        )
        # logger.debug(
        #     f"Computed edge-sample covariance {COV_E_S.shape} in {1e3 * (ttime.monotonic() - start_time):.0f} ms.",
        # )

        # sample-sample upper {i,j}
        i, j = np.triu_indices(self.n_sample, k=1)
        COV_S_S = np.eye(self.n_sample) + self.jitter
        COV_S_S[i, j] = self.callback(
            np.sqrt(
                np.square(self.sample_points[j] - self.sample_points[i]).sum(axis=1),
            ),
            **self.callback_kwargs,
        )
        COV_S_S[j, i] = COV_S_S[i, j]
        COV_S_S += np.diag(COV_MAT_JITTER * np.diag(COV_S_S))  # add some jitter
        # logger.debug(
        #     f"Computed sample-sample covariance {COV_S_S.shape} in {1e3 * (ttime.monotonic() - start_time):.0f} ms.",
        # )

        # this is typically the bottleneck
        inv_COV_S_S = fast_psd_inverse(COV_S_S)
        # logger.debug(
        #     f"Inverted sample-sample covariance {COV_S_S.shape} in {1e3 * (ttime.monotonic() - start_time):.0f} ms.",
        # )

        self.COV_S_S = COV_S_S

        # compute the weights
        self.A = COV_E_S @ inv_COV_S_S

        if (self.A.sum(axis=-1) > 1.0).any():
            raise ValueError(
                f"Propagation operator is unstable (A_max = {self.A.sum(axis=-1).max()}).",
            )

        self.B = np.linalg.cholesky(COV_E_E - self.A @ COV_E_S.T)

        # logger.debug(
        #     f"Computed Cholesky decomposition {COV_E_E.shape} in {duration_ms:.0f} ms.",
        # )

        self.values = np.zeros((self.n_extrusion, self.n_cross_section))

        initial_slice = self.B @ np.random.standard_normal(self.n_cross_section)
        initial_slice *= np.sqrt(np.diag(COV_E_E) / initial_slice.var())

        self.values[:] = initial_slice

        logger.debug(
            f"Computed propagators with shape {self.A.shape} in {humanize_time(ttime.monotonic() - covariance_s)}",
        )

    def run(self):
        if not hasattr(self, "A"):
            self.compute_covariance_matrices()

        n_steps = 2 * self.n_extrusion
        BUFFER = np.random.standard_normal(
            (self.n_extrusion + n_steps, self.n_cross_section),
        )

        iterator = np.arange(n_steps)[::-1]

        for buffer_index in iterator:
            new_values = self.A @ BUFFER[
                buffer_index + self.extrusion_sample_index + 1,
                self.cross_section_sample_index,
            ] + self.B @ np.random.standard_normal(size=self.n_live_edge)
            BUFFER[buffer_index] = new_values

        self.values = da.asarray(BUFFER[: self.n_extrusion])



import os

from maria.utils.io import read_yaml

here, this_filename = os.path.split(__file__)

FRAMES = read_yaml(f"{here}/frame.yml")


def parse_frame(frame):
    for key, config in FRAMES.items():
        if frame in [key, *config["aliases"]]:
            return key
    raise ValueError(f"Invalid frame '{frame}'")


class Frame:
    def __init__(self, frame):
        if isinstance(frame, type(self)):
            frame = frame.name
        self.name = parse_frame(frame)

    def __getattr__(self, key, delim="_"):
        res = FRAMES[self.name]
        while key:
            if key in res:
                return res[key]
            else:
                parts = key.split(delim)
                res = res.get(parts[0], {})
                key = delim.join(parts[1:])
        raise AttributeError(f"'{type(self)}' object has no attribute '{key}'")

    def __repr__(self):
        return f"{type(self).__name__}('{self.name}')"

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other

    def __copy__(self):
        return Frame(self.name)

    def __deepcopy__(self, memo):
        return self.__copy__()



import numpy as np

from ..constants import c, h, k_B  # noqa


def rayleigh_jeans_spectrum(T_RJ: float, nu: float):
    return 2 * k_B * nu**2 * T_RJ / c**2


def inverse_rayleigh_jeans_spectrum(I_nu: float, nu: float):
    return I_nu * c**2 / (2 * k_B * nu**2)


def planck_spectrum(T_b: float, nu: float):
    return 2 * h * nu**3 / (c**2 * np.expm1(h * nu / (k_B * T_b)))


def inverse_planck_spectrum(I_nu: float, nu: float):
    return (h * nu / k_B) / np.log1p(2 * h * nu**3 / (I_nu * c**2))



from __future__ import annotations

import os

import h5py
import numpy as np
import scipy as sp
from jax import scipy as jsp

from ..io import fetch
from ..site import REGIONS, InvalidRegionError, all_regions
from ..units import Quantity

here, this_filename = os.path.split(__file__)


class AtmosphericSpectrum:
    def __init__(self, region, altitude: int = None, source: str = "am", refresh_cache=False):
        if region not in all_regions:
            raise InvalidRegionError(region)

        self.region = region
        self.altitude = altitude or REGIONS.loc[self.region, "altitude"]
        self.source = source

        self.cache_path = fetch(
            f"atmosphere/spectra/{self.source}/v3/{self.region}.h5",
            max_age=30 * 86400,
            refresh=refresh_cache,
        )

        key_mapping = {
            "rayleigh_jeans_temperature_K": "_emission",
            "opacity_nepers": "_opacity",
            "excess_path_m": "_path_delay",
        }

        with h5py.File(self.cache_path, "r") as f:
            # dims are (alt, temp, pwv, el)
            self.side_altitude = f["side_altitude_m"][:].astype(float)
            self.side_base_temperature = f["side_base_temperature_K"][:].astype(float)
            self.side_elevation = np.radians(f["side_elevation_deg"][:].astype(float))
            self.side_zenith_pwv = f["side_zenith_pwv_mm"][:].astype(float)
            self.side_nu = f["side_nu_Hz"][:].astype(float)

            # set the max elevation to 90.1 degrees
            # because sometimes rounded 32-bit types overrun 90 degrees in the interpolator
            self.side_elevation[-1] = np.radians(90.1)

            for key, mapping in key_mapping.items():
                d = f[key][:]

                setattr(
                    self,
                    mapping,
                    sp.interpolate.interp1d(self.side_altitude, d, axis=0)(self.altitude),
                )

    @property
    def nu_min(self):
        return Quantity(self.side_nu.min(), "Hz")

    @property
    def nu_max(self):
        return Quantity(self.side_nu.max(), "Hz")

    def __repr__(self):
        return f"""AtmosphericSpectrum({self.nu_min} - {self.nu_max}):
  region: {self.region}
  altitude: {Quantity(self.altitude, "m")}"""

    def _interpolate_quantity(self, quantity, nu, pwv=None, base_temperature=None, elevation=None):
        pwv = pwv or np.median(self.side_zenith_pwv)
        base_temperature = base_temperature or np.median(self.side_base_temperature)
        elevation = elevation or np.radians(45)

        min_pwv = self.side_zenith_pwv.min()
        max_pwv = self.side_zenith_pwv.max()
        if (np.min(pwv) < min_pwv) or (np.max(pwv) > max_pwv):
            raise ValueError(f"PWV (in mm) must be between {min_pwv} and {max_pwv}.")

        min_elevation = self.side_elevation.min()
        max_elevation = self.side_elevation.max()
        if (np.min(elevation) < min_elevation) or (np.max(elevation) > max_elevation):
            raise ValueError(f"Elevation (in degrees) must be between {min_elevation} and {max_elevation}.")

        min_base_temp = self.side_base_temperature.min()
        max_base_temp = self.side_base_temperature.max()
        if (np.min(base_temperature) < min_base_temp) or (np.max(base_temperature) > max_base_temp):
            raise ValueError(f"Base temperature (in Kelvin) must be between {min_base_temp:.01f} and {max_base_temp:.01f}.")

        return jsp.interpolate.RegularGridInterpolator(
            points=self.points,
            values=getattr(self, f"_{quantity}"),
        )((base_temperature, pwv, elevation, nu))

    @property
    def points(self):
        return (
            self.side_base_temperature,
            self.side_zenith_pwv,
            self.side_elevation,
            self.side_nu,
        )

    def emission(self, nu, pwv=None, base_temperature=None, elevation=None):
        return self._interpolate_quantity(
            "emission",
            nu=nu,
            pwv=pwv,
            base_temperature=base_temperature,
            elevation=elevation,
        )

    def opacity(self, nu, pwv=None, base_temperature=None, elevation=None):
        return self._interpolate_quantity(
            "opacity",
            nu=nu,
            pwv=pwv,
            base_temperature=base_temperature,
            elevation=elevation,
        )

    def path_delay(self, nu, pwv=None, base_temperature=None, elevation=None):
        return self._interpolate_quantity(
            "path_delay",
            nu=nu,
            pwv=pwv,
            base_temperature=base_temperature,
            elevation=elevation,
        )

    def transmission(self, nu, pwv=None, base_temperature=None, elevation=None):
        return np.exp(
            -self.opacity(
                nu,
                pwv=pwv,
                base_temperature=base_temperature,
                elevation=elevation,
            ),
        )



from __future__ import annotations

import numpy as np
import pytest
from maria.band import Band, BandList, all_bands, get_band, parse_band
from maria.errors import FrequencyOutOfBoundsError


def test_band_conventions():
    band1 = Band(center=150e9, width=30e9, NET_RJ=1e-5)
    band2 = {"center": 90e9, "width": 30e9, "NEP": 1e-15}
    band3 = "act/pa5/f150"

    for band in [band1, band2, band3]:
        parse_band(band)

    bl1 = BandList(bands=[band1, band2, band3])

    assert bl1[1].name == "f090"

    bl2 = BandList(bands={"first_band": band1, "second_band": band2, "third_band": band3})

    assert bl2[1].name == "second_band"


def test_frequency_limits():
    caught = False
    try:
        Band(center=90e3, width=20e3)
    except FrequencyOutOfBoundsError:
        caught = True
    assert caught

    caught = False
    try:
        Band(center=90e15, width=20e15)
        caught = False
    except FrequencyOutOfBoundsError:
        caught = True
    assert caught


def test_band_manual():
    nu = np.linspace(120e9, 180e9, 64)
    tau = np.exp(-(((nu - 150e9) / (2 * 10e9)) ** 2))
    b = Band(nu=nu, tau=tau)
    b.plot()


def test_band_list():
    bl = BandList()
    for band_name in all_bands:
        band = get_band(band_name)
        bl.add(band)


def test_noise_conversion():
    my_band = Band(
        center=150e9,  # in Hz
        width=30e9,  # in Hz
        efficiency=0.5,  # in K_RJ
        NET_RJ=1e-5,
        spectrum_kwargs={
            "region": "chajnantor",
            "pwv": 1e1,  # in mm
            "elevation": 90,
        },
    )  # in degrees
    print(my_band)



import logging
import os

from ..units import Quantity

logger = logging.getLogger("maria")
here, this_filename = os.path.split(__file__)


def repr_lat_lon(lat, lon):
    lat_repr = Quantity(abs(lat), "deg").repr("dms") + (" N" if lat > 0 else " S")
    lon_repr = Quantity(abs(lon), "deg").repr("dms") + (" E" if lon > 0 else " W")
    return lat_repr, lon_repr


def repr_phi_theta(phi: float, theta: float, frame: str, join: bool = False):
    qphi = Quantity(phi, "rad")
    qtheta = Quantity(theta, "rad")
    if frame == "az/el":
        res = {"az": qphi.repr("deg"), "el": qtheta.repr("deg")}
    elif frame == "ra/dec":
        res = {"ra": qphi.repr("hms"), "dec": qtheta.repr("dms")}
    elif frame == "galactic":
        res = {"l": qphi.repr("deg"), "b": qtheta.repr("deg")}
    else:
        raise ValueError(f"Invalid frame '{frame}'")

    if join:
        res = (f"{key}: {value}" for key, value in res.items())

    return res



# speed of light (m s^-1)
c = 2.99792458e8

# little g (m s^-2)
g = 9.806651

# Planck's constant (kg m^2 s)
h = 6.62607015e-34

# Reduced Planck's constant (kg m^2 s)
hbar = h / 6.283185307179586

# Boltzmann's constant (kg m^2 s^-2 K^-1)
k_B = 1.380649e-23

# CMB temperature (K)
T_CMB = 2.72548

# equatorial radius of the earth (m)
earth_radius = 6.378137e6

MARIA_MIN_NU_HZ = 1e6  # 1 MHz
MARIA_MAX_NU_HZ = 15e12  # 150 THz

DEFAULT_TIME_FORMAT = "YYYY-MM-DD HH:mm:ss.SSS ZZ"
DEFAULT_BAR_FORMAT = "{l_bar}{bar:16}{r_bar}"



from __future__ import annotations

import numpy as np
import pandas as pd


class ReadOutBLIPs:
    """
    Background limited estimates for AtLAST based of from
    Sean Bryans estimator.
    See https://github.com/atlast-telescope/sens_calc for more info.
    """

    def _run(self):
        self._get_levels()

    def _get_levels(self):
        df = pd.read_csv(
            "./maria/noise/results_of_instrument_simulation.csv",
            index_col=0,
        )
        frq_index = np.argmin(np.abs(self.f_c - df["fmin"]))

        used_b_w = df["fmax"][frq_index] - df["fmin"][frq_index]  # GHz
        NET = df["NET_RJ of a Single Detector"][frq_index]

        self.white_noise_RJ = NET / np.sqrt(self.b_w / used_b_w) * 1e-6  # K s^0.5
        self.pink_noise_RJ = 0.3 * (self.white_noise_RJ / 112.0e-6)

        if self.cal:
            self.white_noise_RJ = self.white_noise_RJ * 0.3 / self.abscal
            self.pink_noise_RJ = self.pink_noise_RJ * 0.3 / self.abscal


class InitNoise(ReadOutBLIPs):
    """
    Estimate initial white and pink noise levels
    """

    def __init__(
        self,
        f_c: np.float32 = 93,  # GHz -- central frequency
        b_w: np.float32 = 52,  # GHz -- band width
        abscal: np.float32 = 1.0,  # Calibration term due to atmospheric absorption
        cal: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.f_c = f_c
        self.b_w = b_w
        self.cal = cal
        self.abscal = abscal
        self._run()



import logging
from collections.abc import Mapping

import arrow
import astropy as ap
import numpy as np
import scipy as sp

from ..coords import Coordinates
from ..errors import NoSuitablePlansError, PointingError
from ..site import Site, get_site
from ..units import Quantity
from ..utils import get_day_hour, great_circle_distance, grouper
from .patterns import parse_scan_kwargs
from .plan import Plan
from .plan_list import PlanList

CONSTRAINT_KEYS = ["az", "el", "hour", "min_sun_distance"]
SIDEREAL_DAY_SECONDS = Quantity(86164.0905, "s")
YEAR_SECONDS = Quantity(31_556_926, "s")


logger = logging.getLogger("maria")


class Planner:
    def __init__(
        self,
        target,
        site: Site,
        constraints: Mapping,
        max_lookahead: float = 2 * 365 * 86400.0,
        start_time: float = None,
    ):
        self.target = target

        if isinstance(site, Site):
            self.site = site
        elif isinstance(site, str):
            self.site = get_site(site_name=site)
        else:
            raise ValueError(
                "'site' must be either a Site object or a string.",
            )

        self.constraints = {}
        if not isinstance(constraints, Mapping):
            raise TypeError("'constraints' must be a dict or a mapping")
        for key, value in constraints.items():
            if key in ["az", "el"]:
                self.constraints[key] = (Quantity(value[0], "deg"), Quantity(value[1], "deg"))
            elif key == "hour":
                self.constraints[key] = value
            elif key == "min_sun_distance":
                self.constraints[key] = Quantity(value, "deg")
            else:
                raise ValueError(f"constraint keys must be one of {CONSTRAINT_KEYS}")

        self.start_time = arrow.get(start_time or arrow.get().timestamp(), tzinfo=self.site.timezone)
        self.max_lookahead = max_lookahead

    @property
    def start_day(self):
        return arrow.get(year=self.start_time.year, month=self.start_time.month, day=self.start_time.day)

    @staticmethod
    def apply_constraint(x, c1, c2):
        if c1 < c2:
            return (x >= c1) & (x <= c2)
        if c1 > c2:
            return (x >= c1) | (x <= c2)

    def generate_obs_intervals(self, total_duration: float, max_chunk_duration: float, test_points_per_chunk: int = 16):
        max_chunk_duration = max_chunk_duration or np.inf

        self.delta_t_test = np.arange(0, self.max_lookahead, max_chunk_duration / test_points_per_chunk)
        self.mask = np.ones_like(self.delta_t_test, dtype=bool)

        if any([key in self.constraints for key in ["az", "el"]]):
            sidereal_day_seconds_test = self.delta_t_test % SIDEREAL_DAY_SECONDS.s
            sidereal_day_seconds_samples = np.linspace(0, SIDEREAL_DAY_SECONDS.s, 256)
            target_sample_t = self.start_time.timestamp() + sidereal_day_seconds_samples
            self.c_target = Coordinates(
                phi=self.target.center[0].radians,
                theta=self.target.center[1].radians,
                t=target_sample_t,
                frame=self.target.frame,
                earth_location=self.site.earth_location,
            )

            self.target_az_test = Quantity(
                sp.interpolate.interp1d(sidereal_day_seconds_samples, np.unwrap(self.c_target.az))(sidereal_day_seconds_test)
                % (2 * np.pi),
                "rad",
            )
            self.target_el_test = Quantity(
                sp.interpolate.interp1d(sidereal_day_seconds_samples, self.c_target.el)(sidereal_day_seconds_test), "rad"
            )

            if "el" in self.constraints:
                self.mask &= self.target_el_test >= self.constraints["el"][0]
                if not self.mask.any():
                    raise PointingError(
                        f"target is never above an elevation of {self.constraints['el'][0]} "
                        f"(el_max = {self.target_el_test.max()})"
                    )
                self.mask &= self.target_el_test <= self.constraints["el"][1]
                if not self.mask.any():
                    raise PointingError(
                        f"target is never below an elevation of {self.constraints['el'][1]} "
                        f"(el_min = {self.target_el_test.min()})"
                    )

            if "az" in self.constraints:
                self.mask &= self.apply_constraint(self.target_az_test, *self.constraints["az"])

        if "min_sun_distance" in self.constraints:
            year_fraction_test = (self.delta_t_test % YEAR_SECONDS.s) / YEAR_SECONDS.s
            year_fraction_samples = np.linspace(0, 1, 256)
            sun_sample_t = self.start_time.timestamp() + year_fraction_samples * YEAR_SECONDS.s

            sun = ap.coordinates.get_sun(ap.time.Time(sun_sample_t, format="unix"))
            sun_distance_samples = great_circle_distance(
                self.target.center[0].rad, self.target.center[1].rad, sun.ra.rad, sun.dec.rad
            )

            self.sun_distance_test = Quantity(
                sp.interpolate.interp1d(year_fraction_samples, sun_distance_samples)(year_fraction_test), "rad"
            )

            self.mask &= self.sun_distance_test >= self.constraints["min_sun_distance"]

        if "hour" in self.constraints:
            day_hour_test = (get_day_hour(self.start_time) + (self.delta_t_test % 86400) / 3600) % 24
            self.mask &= self.apply_constraint(day_hour_test, *self.constraints["hour"])

        logger.debug(f"{1e2 * self.mask.sum() / len(self.mask):.02f}% of times satisfy the observing constraints")

        chunks = []
        duration = 0
        for start, end in grouper(
            self.mask, min_length=test_points_per_chunk, max_length=test_points_per_chunk + 1, overlap=True
        ):
            this_chunk_duration = np.min(
                [max_chunk_duration, self.delta_t_test[end] - self.delta_t_test[start], total_duration - duration]
            )

            debug_str_parts = [f"start = {arrow.get(self.start_time.timestamp() + self.delta_t_test[start])}"]
            if "az" in self.constraints:
                chunk_az = self.target_az_test[start:end]
                debug_str_parts.append(f"az = {chunk_az.mean()} ± {chunk_az.std()}")
            if "el" in self.constraints:
                chunk_el = self.target_el_test[start:end]
                debug_str_parts.append(f"el = {chunk_el.mean()} ± {chunk_el.std()}")
            logger.debug(f"Found good chunk ({', '.join(debug_str_parts)})")

            chunk_start = self.start_time.shift(seconds=self.delta_t_test[start])
            chunks.append({"start_time": chunk_start, "duration": this_chunk_duration})
            duration += this_chunk_duration

            if duration >= total_duration:
                break

        return chunks

    def generate_plans(
        self, total_duration: float, max_chunk_duration: float = 1800, scan_options: Mapping = {}, **plan_kwargs
    ):
        scan_options = parse_scan_kwargs(scan_options, default_radius=self.target.width.deg / 2)

        # scan_options["radius"] = scan_options.get("radius", self.target.width.deg / 2)

        chunks = self.generate_obs_intervals(total_duration=total_duration, max_chunk_duration=max_chunk_duration)
        total_duration_of_chunks = sum([chunk["duration"] for chunk in chunks])

        if not chunks:
            raise NoSuitablePlansError(
                constraints=self.constraints,
                max_chunk_duration=max_chunk_duration,
                tmin=self.start_time,
                tmax=self.start_time.shift(seconds=self.max_lookahead),
            )

        if total_duration_of_chunks < total_duration:
            logger.warning(
                f"Found only {total_duration_of_chunks} seconds of observing time between {self.start_time}"
                f" and {self.start_time.shift(seconds=self.max_lookahead)}; consider increasing the "
                "'max_lookahead' parameter"
            )

        plans = [
            Plan.generate(
                scan_center=self.target.center,
                frame=self.target.frame,
                site=self.site,
                scan_options=scan_options,
                **plan_kwargs,
                **chunk,
            )
            for chunk in chunks
        ]

        return PlanList(plans)

    def generate_plan(self, total_duration: float, **plan_kwargs):
        return self.generate_plans(total_duration=total_duration, max_chunk_duration=total_duration, **plan_kwargs)[0]



from __future__ import annotations

import logging
import os
import time as ttime

import arrow
import numpy as np
from tqdm import tqdm

from ..atmosphere import DEFAULT_ATMOSPHERE_KWARGS, Atmosphere
from ..cmb import CMB, DEFAULT_CMB_KWARGS, generate_cmb, get_cmb
from ..coords import Coordinates
from ..errors import PointingError
from ..instrument import Instrument, get_instrument
from ..io import fetch, humanize_time
from ..map import Map, load
from ..plan import Plan, PlanList, get_plan
from ..site import Site, get_site
from ..tod import TOD
from ..utils import read_yaml
from .atmosphere import AtmosphereMixin
from .base import BaseSimulation
from .cmb import CMBMixin
from .map import MapMixin
from .noise import NoiseMixin
from .observation import Observation

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

MIN_ELEVATION_WARN = 10  # degrees
MIN_ELEVATION_ERROR = 5  # degrees


class InvalidSimulationParameterError(Exception):
    def __init__(self, invalid_keys):
        super().__init__(
            f"The parameters {invalid_keys} are not valid simulation parameters!",
        )


master_params = read_yaml(f"{here}/params.yml")


def parse_sim_kwargs(kwargs, master_kwargs, strict=False):
    parsed_kwargs = {k: {} for k in master_kwargs.keys()}
    invalid_kwargs = {}

    for k, v in kwargs.items():
        parsed = False
        for sub_type, sub_kwargs in master_kwargs.items():
            if k in sub_kwargs.keys():
                parsed_kwargs[sub_type][k] = v
                parsed = True
        if not parsed:
            invalid_kwargs[k] = v

    if len(invalid_kwargs) > 0:
        if strict:
            raise InvalidSimulationParameterError(
                invalid_keys=list(invalid_kwargs.keys()),
            )

    return parsed_kwargs


class Simulation(AtmosphereMixin, CMBMixin, MapMixin, NoiseMixin):
    """
    A simulation of a telescope. This is what users should touch, primarily.
    """

    @classmethod
    def from_config(cls, config: dict = {}, **params):
        return cls(**{**config, **params})

    def __init__(
        self,
        instrument: Instrument | str,
        plans: PlanList | list[Plan | str],
        site: Site | str,
        atmosphere: Atmosphere | str = None,
        atmosphere_kwargs: dict = {},
        cmb: CMB | str = None,
        cmb_kwargs: dict = {},
        map: Map | str = None,
        map_kwargs: dict = {},
        noise: bool = True,
        noise_kwargs: dict = {},
        progress_bars: bool = True,
        keep_mean_signal: bool = False,
        dtype: type = np.float32,
    ):
        self.atmosphere = atmosphere
        self.atmosphere_kwargs = DEFAULT_ATMOSPHERE_KWARGS.copy()
        self.atmosphere_kwargs.update(atmosphere_kwargs)

        self.map_kwargs = map_kwargs

        self.noise = noise
        self.noise_kwargs = noise_kwargs

        self.dtype = dtype
        self.disable_progress_bars = (not progress_bars) or (logging.getLevelName(logger.level) == "DEBUG")

        instrument_init_s = ttime.monotonic()

        # parsed_sim_kwargs = parse_sim_kwargs(kwargs, master_params)

        if isinstance(instrument, Instrument):
            self.instrument = instrument
        else:
            self.instrument = get_instrument(name=instrument)

        logger.debug(f"Initialized instrument in {humanize_time(ttime.monotonic() - instrument_init_s)}.")
        site_init_s = ttime.monotonic()

        if isinstance(site, Site):
            self.site = site
        elif isinstance(site, str):
            self.site = get_site(site_name=site)
        else:
            raise ValueError(
                "'site' must be either a Site object or a string.",
            )

        logger.debug(f"Initialized site in {humanize_time(ttime.monotonic() - site_init_s)}.")
        plan_init_s = ttime.monotonic()

        if isinstance(plans, str):
            plans = [get_plan(plan_name=plans)]
        elif isinstance(plans, Plan):
            plans = [plans]
        elif isinstance(plans, PlanList):
            plans = plans
        elif not isinstance(plans, list):
            raise TypeError("plans must be a plan or a list of plans")

        self.plans = PlanList(plans)

        logger.debug(f"Initialized plans in {humanize_time(ttime.monotonic() - plan_init_s)}.")

        self.obs_list = []
        for plan in self.plans:
            obs_weather_kwargs = self.atmosphere_kwargs.pop("weather") if "weather" in self.atmosphere_kwargs else {}
            obs = Observation(
                instrument=self.instrument,
                plan=plan,
                site=self.site,
                atmosphere=self.atmosphere,
                atmosphere_kwargs=self.atmosphere_kwargs,
                weather_kwargs=obs_weather_kwargs,
            )

            self.obs_list.append(obs)

        if cmb:
            cmb_start_s = ttime.monotonic()
            self._init_cmb(cmb, **cmb_kwargs)
            logger.debug(f"Initialized CMB simulation in {humanize_time(ttime.monotonic() - cmb_start_s)}.")

        if map:
            map_start_s = ttime.monotonic()
            self._initialize_map(map, **map_kwargs)
            logger.debug(f"Initialized map simulation in {humanize_time(ttime.monotonic() - map_start_s)}.")

        if noise:
            noise_start_s = ttime.monotonic()
            logger.debug(f"Initialized noise simulation in {humanize_time(ttime.monotonic() - noise_start_s)}.")

        #     # self.start = arrow.get(self.boresight.t.min()).to("utc")
        #     # self.end = arrow.get(self.boresight.t.max()).to("utc")

        #     if atmosphere:
        #         atmosphere_init_start_s = ttime.monotonic()

        #         # give it the observation, so that it knows about pointing, site, etc. (kind of cursed)
        #         obs.atmosphere.initialize(obs)

        #         logger.debug(
        #             f"Initialized atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_init_start_s)}."
        #         )

        # logger.debug(f"Initialized simulation in {humanize_time(ttime.monotonic() - sim_start_s)}.")

    def run(self, units: str = "K_RJ"):
        tods = []
        for obs_index, obs in enumerate(self.obs_list):
            logger.info(f"Simulating observation {obs_index + 1} of {len(self.obs_list)}")
            obs_start_s = ttime.monotonic()
            tods.append(self.run_obs(obs).to(units))
            logger.info(
                f"Simulated observation {obs_index + 1} of {len(self.obs_list)} "
                f"in {humanize_time(ttime.monotonic() - obs_start_s)}"
            )
        return tods

    def run_obs(self, obs: Observation) -> TOD:
        obs.loading = {}

        if hasattr(obs, "atmosphere"):
            atmosphere_sim_start_s = ttime.monotonic()
            obs.atmosphere.initialize(instrument=obs.instrument, boresight=obs.boresight, site=obs.site)
            self._simulate_atmosphere(obs)
            obs.loading["atmosphere"] = self._compute_atmospheric_loading(obs)
            logger.debug(f"Ran atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_sim_start_s)}.")

        if hasattr(self, "cmb"):
            cmb_sim_start_s = ttime.monotonic()
            obs.loading["cmb"] = self._compute_cmb_loading(obs)
            logger.debug(f"Ran CMB simulation in {humanize_time(ttime.monotonic() - cmb_sim_start_s)}.")

        if hasattr(self, "map"):
            map_sim_start_s = ttime.monotonic()
            self._sample_maps(obs)
            logger.debug(f"Ran map simulation in {humanize_time(ttime.monotonic() - map_sim_start_s)}.")

        # number of bands are lost here
        if self.noise:
            noise_sim_start_s = ttime.monotonic()
            self._simulate_noise(obs)
            logger.debug(f"Ran noise simulation in {humanize_time(ttime.monotonic() - noise_sim_start_s)}.")

        gain_error = np.exp(
            obs.instrument.dets.gain_error * np.random.standard_normal(size=obs.instrument.dets.n),
        )

        for field in obs.loading:
            if field in ["noise"]:
                continue

            obs.loading[field] *= gain_error[:, None]

        metadata = {
            "atmosphere": False,
            "sim_time": arrow.now(),
            "altitude": float(obs.site.altitude.m),
            "region": obs.site.region,
        }

        if hasattr(obs, "atmosphere"):
            metadata["atmosphere"] = True
            metadata["pwv"] = float(np.round(obs.atmosphere.weather.pwv, 3))
            metadata["base_temperature"] = float(np.round(obs.atmosphere.weather.temperature[0], 3))
        else:
            metadata["atmosphere"] = False

        return TOD(
            data=obs.loading,
            dets=obs.instrument.dets,
            coords=obs.coords,
            units="pW",
            metadata=metadata,
        )

    def plot_counts(self, x_bins=100, y_bins=100):
        self.plan.plot_counts(instrument=self.instrument, x_bins=x_bins, y_bins=y_bins)

    def __repr__(self):
        instrument_tree = "├ " + self.instrument.__repr__().replace("\n", "\n│ ")
        site_tree = "├ " + self.site.__repr__().replace("\n", "\n│ ")
        plan_tree = "├ " + self.plans.__repr__().replace("\n", "\n│ ")

        trees = []
        attrs = [attr for attr in ["atmosphere", "cmb", "map"] if getattr(self, attr, None)]
        if attrs:
            for attr in attrs[:-1]:
                trees.append("├ " + getattr(self, attr).__repr__().replace("\n", "\n│ "))
            trees.append("└ " + getattr(self, attrs[-1]).__repr__().replace("\n", "\n  "))

        trees_string = "\n".join(trees)

        return f"""Simulation
{instrument_tree}
{site_tree}
{plan_tree}
{trees_string}"""

    @property
    def total_loading(self):
        return sum([d for d in self.loading.values()])

    @property
    def min_time(self):
        return self.obs_list[0].plan.start_time

    @property
    def max_time(self):
        return self.obs_list[-1].plan.end_time



from __future__ import annotations

import logging

import healpy as hp
import numpy as np
import pandas as pd
import scipy as sp

from ..io import fetch
from ..map import ProjectionMap
from ..units import Quantity
from .cmb import CMB

# shut up healpy I don't care about the resolution
logging.getLogger("healpy").setLevel(logging.WARNING)

CMB_SPECTRUM_CACHE_PATH = "/tmp/maria-data/cmb/spectrum.txt"
CMB_SPECTRUM_CACHE_MAX_AGE = 30 * 86400  # one month

CMB_SOURCES = {
    "planck": {"spectrum": "cmb/spectra/planck.csv"},
    "camb": {"spectrum": "cmb/spectra/camb.csv"},
}


def get_cmb_spectrum(source="camb"):
    cmb_spectrum_path = fetch(
        CMB_SOURCES[source]["spectrum"],
        max_age=CMB_SPECTRUM_CACHE_MAX_AGE,
    )

    return pd.read_csv(cmb_spectrum_path, index_col=0)


def generate_cmb(nside: int = 2048, seed=123456, **kwargs):
    """
    Generate a new CMB.

    Taken from https://www.zonca.dev/posts/2020-09-30-planck-spectra-healpy.html
    """

    cl = get_cmb_spectrum(source="planck")
    lmax = cl.ell.max()

    alm = hp.synalm((cl.TT, cl.EE, cl.BB, cl.TE), lmax=lmax, new=True)
    cmb_data = hp.alm2map(alm, nside=nside, lmax=lmax)

    return CMB(
        data=cmb_data[:, None, None, :],
        stokes="IQU",
        units="K_CMB",
        frame="galactic",
        nu=148e9,
    )


def generate_cmb_patch(
    width: float = 5.0,
    height: float = None,
    center: tuple[float, float] = None,
    resolution: float = None,
    frame: str = "ra/dec",
    degrees: bool = True,
    buffer: int = 2,
):
    width = Quantity(width, "deg" if degrees else "rad").rad
    height = Quantity(height, "deg" if degrees else "rad").rad if height is not None else width
    resolution = Quantity(resolution, "deg" if degrees else "rad").rad if resolution is not None else width / 256
    center = Quantity(center, "deg" if degrees else "rad").rad if center is not None else (0.0, 0.0)

    nx = int(width / resolution)
    ny = int(height / resolution)

    dx = width / nx
    dy = height / ny

    x_buffer_pixels = buffer * nx
    y_buffer_pixels = buffer * ny

    nx_gen = 2 * x_buffer_pixels + nx
    ny_gen = 2 * y_buffer_pixels + ny

    kx = np.fft.fftfreq(nx_gen, dx)
    ky = np.fft.fftfreq(ny_gen, dy)
    KY, KX = np.meshgrid(ky, kx)
    K = np.sqrt(KX**2 + KY**2)

    cmb_spectrum = get_cmb_spectrum()
    PS = sp.interpolate.interp1d(cmb_spectrum.ell.values, cmb_spectrum.TT.values, bounds_error=False, fill_value=0)(
        2 * np.pi * K
    )

    PS /= resolution**2  # convert from uK^2 radians^2 to uK^2 pixels^2

    complex_map = np.fft.ifft2(np.sqrt(PS) * np.fft.fft2(np.random.standard_normal((ny_gen, nx_gen))))
    m = complex_map.real[:nx, :ny]

    return ProjectionMap(
        data=(m - m.mean())[None], center=center, width=width, nu=148e9, frame=frame, units="K_CMB", degrees=False
    )



from __future__ import annotations

import os

import maria
import numpy as np
import pytest
from maria import Simulation, all_instruments, all_sites, get_plan
from maria.mappers import BinMapper
from maria.utils import read_yaml

here, this_filename = os.path.split(__file__)

# all_test_plan_configs = list(read_yaml(f"{here}/../configs/test_plans.yml").values())

all_instruments.pop(all_instruments.index("alma/ALMA"))

n_sims = 10
test_instruments = np.random.choice(a=all_instruments, size=n_sims)
test_sites = np.random.choice(a=all_sites, size=n_sims)
test_az = np.random.uniform(low=0, high=360, size=n_sims)
test_el = np.random.uniform(low=30, high=90, size=n_sims)


@pytest.mark.parametrize(
    "instrument,site,az,el",
    zip(test_instruments, test_sites, test_az, test_el),
)
def test_pipeline(instrument, site, az, el):
    site = maria.get_site(site)

    plan = maria.Plan.generate(
        scan_pattern="daisy",
        scan_options={"radius": 0.5, "speed": 0.1},  # in degrees
        duration=10,  # in seconds
        sample_rate=15,  # in Hz
        scan_center=(az, el),
        frame="az/el",
    )

    sim = Simulation(
        instrument=instrument, site=site, plans=[plan], atmosphere="2d", cmb="generate", cmb_kwargs={"nside": 256}
    )

    tod = sim.run()[0]

    for field in ["atmosphere", "cmb"]:
        if np.isnan(tod.data[field]).any():
            raise ValueError(f"There are NaNs in the '{field}' field.")

    tod = tod.to("K_RJ")

    mapper = BinMapper(
        tod_preprocessing={
            "remove_spline": {"knot_spacing": 30, "remove_el_gradient": True},
            "remove_modes": {"modes_to_remove": 1},
        },
        map_postprocessing={
            "gaussian_filter": {"sigma": 1},
        },
        units="mK_RJ",
        tods=[tod],
    )

    output_map = mapper.run()

    assert output_map.weight.sum() > 0



from __future__ import annotations

import glob
import os
from typing import Mapping

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.collections import EllipseCollection
from matplotlib.patches import Patch

from ..array import Array, ArrayList, get_array_config  # noqa
from ..band import BAND_CONFIGS, Band, BandList, parse_band  # noqa
from ..beam import compute_angular_fwhm
from ..units import Quantity
from ..utils import HEX_CODE_LIST, flatten_config, get_rotation_matrix_2d, read_yaml  # noqa

here, this_filename = os.path.split(__file__)

INSTRUMENT_CONFIGS = {}
for path in glob.glob(f"{here}/configs/*.yml"):
    key = os.path.split(path)[1].split(".")[0]
    INSTRUMENT_CONFIGS[key] = read_yaml(path)
INSTRUMENT_CONFIGS = flatten_config(INSTRUMENT_CONFIGS)

# better formatting for pandas dataframes
# pd.set_eng_float_format()

for name, config in INSTRUMENT_CONFIGS.items():
    config["aliases"] = config.get("aliases", [])
    config["aliases"].append(name.lower())

INSTRUMENT_DISPLAY_COLUMNS = [
    "description",
    # "field_of_view",
    # "primary_size",
    # "bands",
]


# def get_instrument_config(instrument_name=None, **kwargs):
#     if instrument_name not in INSTRUMENT_CONFIGS.keys():
#         raise InvalidInstrumentError(instrument_name)
#     instrument_config = INSTRUMENT_CONFIGS[instrument_name].copy()
#     return instrument_config


# def get_instrument(instrument_name="default", **kwargs):
#     """
#     Get an instrument from a pre-defined config.
#     """
#     if instrument_name:
#         for key, config in INSTRUMENT_CONFIGS.items():
#             if instrument_name.lower() in config.get("aliases", []):
#                 instrument_name = key
#         if instrument_name not in INSTRUMENT_CONFIGS.keys():
#             raise InvalidInstrumentError(instrument_name)
#         instrument_config = INSTRUMENT_CONFIGS[instrument_name].copy()
#     else:
#         instrument_config = {}
#     instrument_config.update(kwargs)
#     return Instrument.from_config(instrument_config)


subarray_params_to_inherit = [
    "array_packing",
    "array_shape",
    "array_offset",
    "beam_spacing",
    "max_baseline",
    "baseline_packing",
    "baseline_shape",
    "baseline_offset",
    "bath_temp",
    "polarization",
    "primary_size",
    "bands",
    "field_of_view",
]

band_params_to_inherit = [
    "time_constant",  # seconds
    "white_noise",  # Kelvin
    "pink_noise",  # Kelvin / s
    "efficiency",
]

passband_params_to_inherit = {
    "center": 150,  # GHz
    "width": 30,  # GHz
    "shape": "top_hat",
}

allowed_subarray_params = {
    "n": "int",
    "array_packing": "float",
    "array_shape": "float",
    "array_offset": "float",
    "beam_spacing": "float",
    "max_baseline": "float",
    "baseline_packing": "float",
    "baseline_shape": "float",
    "baseline_offset": "float",
    "bath_temp": "float",
    "polarization": "float",
    "primary_size": "float",
    "field_of_view": "float",
    "bands": "float",
}


def get_instrument(name=None, **kwargs):
    config = get_instrument_config(name) if name else {}
    config.update(kwargs)
    return Instrument.from_config(config)


def get_instrument_config(name):
    for v in INSTRUMENT_CONFIGS.values():
        if name.lower() in v["aliases"]:
            return v.copy()
    raise KeyError(f"'{name}' is not a valid array name.")


class Instrument:
    @classmethod
    def from_config(cls, config):
        c = config.copy()

        if "array" in c:
            c["arrays"] = [c.pop("array")]

        for key in ["aliases"]:
            if key in c:
                c.pop(key)

        return cls(**c)

    def __init__(
        self,
        arrays: ArrayList | list | dict,
        description: str = "An instrument.",
        documentation: str = "",
        az_vel_limit: float = 1e2,  # in deg/s
        acc_limit: float = 1e2,  # in deg/s^2
    ):
        """
        Parameters
        ----------
        az_vel_limit : type
            The maximum angular speed of the array.
        """

        # if isinstance(arrays, list):
        #     array_index = 1
        #     for array in arrays:
        #         if "name" not in array:
        #             array["name"] = f"array{array_index}"
        #             array_index += 1

        # array_list = []
        # for array in arrays:
        #     if isinstance(array, Mapping):
        #         array_config = get_array_config(**array)
        #         array_list.append(Array.from_config(array_config))
        #     elif isinstance(array, Array):
        #         array_list.append(array)

        # if isinstance(arrays, ArrayList):
        #     self.arrays = arrays.arrays
        # else:
        #     if isinstance(arrays, list):
        #         array_names = [array.name for i in range(len(arrays))]
        #         array_values = arrays

        #     elif isinstance(arrays, dict):
        #         array_names = list(arrays.keys())
        #         array_values = list(arrays.values())
        #     else:
        #         raise ValueError("'arrays' must be a list or a dict.")

        #     self.arrays = []
        #     for array_name, array in zip(array_names, array_values):
        #         if isinstance(array, Array):
        #             array.name = array.name or array_name
        #             self.arrays.append(array)
        #         elif isinstance(array, dict):
        #             if "name" not in array:
        #                 array["name"] = array_name
        #             self.arrays.append(Array.from_config(array))
        #         elif isinstance(array, str):
        #             self.arrays.append(Array.from_kwargs(name=array_name, key=array))

        self.arrays = ArrayList(arrays)
        self.description = description
        self.documentation = documentation
        self.az_vel_limit = az_vel_limit
        self.acc_limit = acc_limit

        # self.primary_size = float(self.dets.primary_size.max())
        # self.field_of_view = np.round(np.degrees(lazy_diameter(self.dets.offsets)), 3)
        # self.baseline = np.round(lazy_diameter(self.dets.baselines), 3)

        # if self.field_of_view < 0.5 / 60:
        #     self.units = "arcsec"
        # elif self.field_of_view < 0.5:
        #     self.units = "arcmin"
        # else:
        #     self.units = "degrees"

    def __repr__(self):
        band_summary = self.arrays.bands.summary()
        band_summary.loc[:, "FWHM"] = [Quantity(self.dets(band=b.name).fwhm.mean(), "rad") for b in self.bands]

        arrays_repr = "\n".join([f"│  {s}" for s in str(self.arrays.summary().__repr__()).split("\n")])
        bands_repr = "\n".join([f"   {s}" for s in str(band_summary.__repr__()).split("\n")])

        s = f"""Instrument({len(self.arrays)} array{"s" if len(self.arrays) > 1 else ""})
├ arrays:
{arrays_repr}
│ 
└ bands:
{bands_repr}"""

        return s

    @property
    def dets(self):
        return self.arrays.combine()

    @property
    def bands(self):
        return self.arrays.bands

    @property
    def sky_x(self):
        return self.dets.sky_x

    @property
    def sky_y(self):
        return self.dets.sky_y

    @property
    def offsets(self):
        return np.c_[self.sky_x, self.sky_y]

    @property
    def baseline_x(self):
        return self.dets.baseline_x

    @property
    def baseline_y(self):
        return self.dets.baseline_y

    @property
    def baseline_z(self):
        return self.dets.baseline_z

    @property
    def baselines(self):
        return np.c_[self.baseline_x, self.baseline_y, self.baseline_z]

    @staticmethod
    def beam_profile(r, fwhm):
        return np.exp(np.log(0.5) * np.abs(r / fwhm) ** 8)

    @property
    def n(self):
        return self.dets.n

    # def angular_beam_filter(self, z, res, beam_profile=None, buffer=1):  # noqa F401
    #     """
    #     Angular beam width (in radians) as a function of depth (in meters)
    #     """
    #     return construct_beam_filter(self.angular_fwhm(z), res, beam_profile=beam_profile, buffer=buffer)

    # def physical_beam_filter(self, z, res, beam_profile=None, buffer=1):  # noqa F401
    #     """
    #     Angular beam width (in radians) as a function of depth (in meters)
    #     """
    #     return construct_beam_filter(self.physical_fwhm(z), res, beam_profile=beam_profile, buffer=buffer)

    def plot(self, z=np.inf, plot_pol_angles=False):
        self.dets.plot(z=z, plot_pol_angles=plot_pol_angles)


instrument_data = pd.DataFrame(INSTRUMENT_CONFIGS).reindex(INSTRUMENT_DISPLAY_COLUMNS).T

# for instrument_name, config in INSTRUMENT_CONFIGS.items():
#     instrument = get_instrument(instrument_name)
#     f_list = sorted(np.unique([band.center for band in instrument.dets.bands]))
#     instrument_data.at[instrument_name, "f [GHz]"] = "/".join([str(f) for f in f_list])
#     instrument_data.at[instrument_name, "n"] = instrument.dets.n

all_instruments = list(instrument_data.index)


class InvalidInstrumentError(Exception):
    def __init__(self, invalid_instrument):
        super().__init__(
            f"The instrument '{invalid_instrument}' is not supported. "
            f"Supported instruments are:\n\n{instrument_data.__repr__()}",
        )



from __future__ import annotations

from functools import partial

import jax
from jax import numpy as jnp


@partial(jax.jit, static_argnames=["shape", "sample_rate", "knee", "beta", "seed"])
def generate_noise_with_knee(
    shape: tuple,
    sample_rate: float = 1e0,
    knee: float = 0,
    beta: float = 1.0,
    seed: int = 12345,
):
    """
    Simulate white noise for a given time and NEP.
    """

    noise = jax.random.normal(key=jax.random.key(seed), shape=shape) * jnp.sqrt(sample_rate)

    # pink noise
    if knee > 0:
        f = jnp.fft.fftfreq(n=shape[-1], d=1 / sample_rate)
        a = knee / 2
        pink_noise_power_spectrum = jnp.where(f != 0, a / (jnp.abs(f) ** beta), 0)

        weights = jnp.sqrt(2 * sample_rate * pink_noise_power_spectrum)
        noise += jnp.real(
            jnp.fft.ifft(
                weights * jnp.fft.fft(jax.random.normal(key=jax.random.key(seed), shape=shape)),
            ),
        )

    return noise



from __future__ import annotations

import maria
import matplotlib.pyplot as plt
import numpy as np
import pytest
from maria.plan import scan_patterns
from maria.plan.patterns import parse_scan_kwargs


@pytest.mark.parametrize("scan_pattern", scan_patterns.index)
def test_pattern(scan_pattern):
    plan = maria.Plan.generate(scan_pattern=scan_pattern)
    print(plan)

    plan.plot()
    plan.plot_counts()

    plt.close("all")


@pytest.mark.parametrize("scan_pattern", scan_patterns.index)
def test_pattern_speed(scan_pattern):
    #     plan = maria.Plan.generate(scan_pattern=scan_pattern)
    #     print(plan)

    #     plan.plot()
    #     plan.plot_counts()

    #     from maria.plan import scan_patterns

    time = np.arange(0, 3600, 0.01)

    # for index, entry in scan_patterns.iterrows():

    if scan_pattern in ["stare"]:
        return

    for trial in range(16):
        scan_kwargs = {
            "x_throw": np.random.choice(np.geomspace(1e-1, 1e0, 256)),  # in degrees
            "speed": np.random.choice(np.geomspace(1e-1, 1e0, 256)),  # in degrees
        }

        x, y = scan_patterns.loc[scan_pattern].generator(time, **parse_scan_kwargs(scan_kwargs))
        vx = np.diff(x) / np.diff(time)
        vy = np.diff(y) / np.diff(time)

        max_speed = np.sqrt(vx**2 + vy**2).max()

        assert np.isclose(max_speed, scan_kwargs["speed"], rtol=2e-1)



from __future__ import annotations

import os

import numpy as np
import pandas as pd

here, this_filename = os.path.split(__file__)

PREFIXES = pd.read_csv(f"{here}/prefixes.csv", index_col=0)
PREFIXES.loc[""] = "", "", "", 0, 1e0
PREFIXES.sort_values("factor", ascending=True, inplace=True)
PREFIXES.loc[:, "primary"] = np.log10(PREFIXES.factor.values) % 3 == 0



# for a given cross-section, extrude points orthogonally using a callback to compute covariance
from __future__ import annotations

import logging
from typing import Callable

import numpy as np
import pandas as pd
import scipy as sp
from pandas import DataFrame

from ..coords import Coordinates
from ..instrument import Instrument
from ..site import Site
from ..weather import Weather

logger = logging.getLogger("maria")


MIN_RES = {"2d": 2, "3d": 15}
MIN_RES_PER_BEAM = {"2d": 0.1, "3d": 0.5}
MIN_RES_PER_FOV = {"2d": 0.02, "3d": 0.1}

COV_MAT_JITTER = 1e-6


def generate_layers(
    instrument: Instrument,
    boresight: Coordinates,
    weather: Weather,
    site: Site,
    mode: str = "2d",
    angular: bool = True,
    max_height: float = 2e3,  # in meters
    min_res: float = None,  # in meters
    min_res_per_beam: float = None,
    min_res_per_fov: float = None,
    layer_spacing: float = 500,
    pwv_rms_frac: float = 3e-2,
) -> DataFrame:
    """
    Generate atmospheric layers.
    N.B.: We implicitly parametrize this model with a single, constant elevation.
    """

    min_res = min_res or MIN_RES[mode]
    min_res_per_beam = min_res_per_beam or MIN_RES_PER_BEAM[mode]
    min_res_per_fov = min_res_per_fov or MIN_RES_PER_FOV[mode]

    min_el = boresight.el.min()

    h_samples = np.arange(0.0, 20000.0, 1e0)
    z_samples = h_samples / np.sin(min_el)

    fov = instrument.dets.field_of_view
    fwhm = instrument.dets.one_detector_from_each_band().physical_fwhm(z_samples[:, None] + 1e-16).min(axis=1)
    r1 = min_res * np.ones(len(z_samples))
    r2 = min_res_per_beam * fwhm.m
    r3 = min_res_per_fov * z_samples * fov.rad
    res_samples = np.minimum(1e3, np.maximum.reduce([r1, r2, r3]))

    def res_func(h):
        return sp.interpolate.interp1d(h_samples, res_samples)(h)

    if mode == "2d":
        h_boundaries = np.array([0.0, 500.0, 1000.0, 1500.0, 2000.0, 3000.0, 5000.0, 8000.0, 12000.0])
        # h_boundaries = layer_boundaries or np.arange(0, max_height + layer_spacing, layer_spacing)
        process_index = np.arange(len(h_boundaries) - 1)
    elif mode == "3d":
        h_boundaries = [0]
        while True:
            new_h = h_boundaries[-1] + res_func(h_boundaries[-1])
            if new_h > max_height:
                break
            h_boundaries.append(h_boundaries[-1] + res_func(h_boundaries[-1]))
        h_boundaries = np.array(h_boundaries)
        process_index = 0

    h_centers = (h_boundaries[1:] + h_boundaries[:-1]) / 2

    weather_values = weather(altitude=site.altitude.m + h_centers)

    layers = pd.DataFrame(weather_values)
    layers.insert(0, "process_index", process_index)
    layers.insert(1, "h", h_centers)
    layers.insert(2, "dh", np.diff(h_boundaries))
    layers.insert(3, "res", res_func(layers.h))
    layers.insert(4, "z", h_centers / np.sin(min_el))
    layers.insert(5, "angular", angular)

    h_boundaries = [0, *(layers.h.values[:-1] + layers.h.values[1:]) / 2, 1e5]

    for layer_index, (h1, h2) in enumerate(zip(h_boundaries[:-1], h_boundaries[1:])):
        dummy_h = site.altitude.m + np.linspace(h1, h2, 1024)
        h = weather.altitude
        w = weather.absolute_humidity
        total_water = np.trapezoid(np.interp(dummy_h, h, w), x=dummy_h)
        layers.loc[layer_index, "total_water"] = total_water

    def boundary_layer_profile(h, h_0: float = 1e3, alpha: float = 1 / 7):
        return np.exp(-h / h_0) * h**alpha

    rel_var = boundary_layer_profile(layers.h.values) ** 2
    pwv_var = (weather.pwv * pwv_rms_frac) ** 2 * rel_var / sum(rel_var)
    layers.loc[:, "pwv_rms"] = np.sqrt(pwv_var)

    if angular:
        layers.res /= layers.z

    return layers


def construct_extrusion_layers(
    points: float,
    res_func: Callable,
    z_min: float,
    z_max: float,
    mode: str = "3d",
    **mode_kwargs,
):
    triangulation = sp.spatial.Delaunay(points[..., 1:])
    layers = pd.DataFrame(columns=["x", "z", "n", "res", "indices"])

    layer_spacing = 500

    n = 0
    z = z_min if mode == "3d" else layer_spacing / 2

    while z < z_max:
        i = len(layers)

        res = res_func(z)

        # find a line
        wide_lp_x_dense = np.arange(points[..., 1].min(), points[..., 1].max(), 1e0)
        wide_lp_z_dense = z * np.ones(len(wide_lp_x_dense))
        wide_lp_dense = np.c_[wide_lp_x_dense, wide_lp_z_dense]
        interior = triangulation.find_simplex(wide_lp_dense) > -1
        lp_dense_x = wide_lp_x_dense[interior]
        n_lp = np.maximum(2, int(np.ptp(np.atleast_1d(lp_dense_x)) / res))
        lp_x = np.linspace(lp_dense_x.min() - 2 * res, lp_dense_x.max() + 2 * res, n_lp)

        layers.loc[i, "x"] = lp_x
        layers.loc[i, "z"] = z
        layers.loc[i, "n"] = n_lp
        layers.loc[i, "res"] = res
        layers.loc[i, "indices"] = n + np.arange(n_lp)

        z += res if mode == "3d" else layer_spacing
        n += n_lp

    cross_section_x = np.concatenate(layers.x.values)
    cross_section_z = np.concatenate(
        [entry.z * np.ones(entry.n) for index, entry in layers.iterrows()],
    )
    cross_section_points = np.concatenate(
        [cross_section_x[..., None], cross_section_z[..., None]],
        axis=-1,
    )

    extrusion_points = np.arange(
        points[..., 0].min() - 2 * layers.res.min(),
        points[..., 0].max() + 2 * layers.res.min(),
        layers.res.min(),
    )

    return layers, cross_section_points, extrusion_points



from __future__ import annotations

from ..constants import MARIA_MAX_NU_HZ, MARIA_MIN_NU_HZ
from ..units import Quantity
from .plan import NoSuitablePlansError  # noqa


class PointingError(Exception): ...


class ShapeError(Exception): ...


class ConfigurationError(Exception): ...


class FrequencyOutOfBoundsError(Exception):
    def __init__(self, nu=None, center_and_width=None):
        qmin_nu = Quantity(MARIA_MIN_NU_HZ, units="Hz")
        qmax_nu = Quantity(MARIA_MAX_NU_HZ, units="Hz")

        if nu:
            super().__init__(
                f"Bad frequencies nu={Quantity(nu, 'Hz')}; maria supports frequencies between {qmin_nu} and {qmax_nu}."
            )

        if center_and_width:
            center, width = center_and_width
            super().__init__(
                f"Band with center {Quantity(center, 'Hz')} and width {Quantity(width, 'Hz')} has a passband extending "
                f"out of bounds; maria supports frequencies between {qmin_nu} and {qmax_nu}."
            )



from __future__ import annotations

import logging
import pathlib
from collections.abc import Mapping

import astropy as ap
import h5py
import pandas as pd
import yaml

logger = logging.getLogger("maria")


def flatten_config(m: dict, prefix: str = ""):
    """
    Turn any dict into a mapping of mappings.
    """

    # if too shallow, add a dummy index
    if not isinstance(m, Mapping):
        return flatten_config({"": m}, prefix="")

    # if too shallow, add a dummy index
    if not all(isinstance(v, Mapping) for v in m.values()):
        return flatten_config({"": m}, prefix="")

    # recursion!
    items = []
    for k, v in m.items():
        new_key = f"{prefix}/{k}" if prefix else k
        if all(isinstance(vv, Mapping) for vv in v.values()):
            items.extend(flatten_config(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)


def read_yaml(path: str):
    """
    Return a YAML file as a dict
    """
    res = yaml.safe_load(pathlib.Path(path).read_text())
    return res if res is not None else {}


def test_file(path) -> bool:
    ext = path.split(".")[-1]
    try:
        if ext in ["h5"]:
            with h5py.File(path, "r") as f:
                f.keys()
        elif ext in ["csv"]:
            pd.read_csv(path)
        elif ext in ["txt", "dat"]:
            with open(path) as f:
                f.read()
        elif ext in ["fits"]:
            ap.io.fits.open(path)
    except Exception:
        return False

    return True



from __future__ import annotations

import logging
import os
import time as ttime
from collections.abc import Sequence

import numpy as np
from tqdm import tqdm

from ..coords import infer_center_width_height
from ..io import humanize_time, repr_phi_theta
from ..map import ProjectionMap
from ..tod import TOD
from ..units import Quantity
from .base import BaseProjectionMapper

np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class BinMapper(BaseProjectionMapper):
    def __init__(
        self,
        tods: Sequence[TOD],
        center: tuple[float, float] = None,
        stokes: str = None,
        width: float = None,
        height: float = None,
        resolution: float = None,
        frame: str = "ra/dec",
        units: str = "K_RJ",
        degrees: bool = True,
        min_time: float = None,
        max_time: float = None,
        timestep: float = None,
        tod_preprocessing: dict = {},
        map_postprocessing: dict = {},
        progress_bars: bool = True,
    ):
        super().__init__(
            tods=tods,
            stokes=stokes,
            center=center,
            width=width,
            height=height,
            resolution=resolution,
            frame=frame,
            units=units,
            tod_preprocessing=tod_preprocessing,
            map_postprocessing=map_postprocessing,
            min_time=min_time,
            max_time=max_time,
            timestep=timestep,
            degrees=degrees,
            progress_bars=progress_bars,
        )

    def _run(self):
        """
        The actual mapper for the BinMapper.
        """

        map_sum = np.zeros(self.map_size)
        map_wgt = np.zeros(self.map_size)

        pbar = tqdm(
            self.tods,
            total=len(self.tods),
            desc=f"Mapping",
            postfix={"tod": f"1/{len(self.tods)}"},
        )

        for tod in pbar:
            if not tod.shape[0] > 0:
                continue

            P = self.map.pointing_matrix(coords=tod.coords, dets=tod.dets)
            D = tod.signal.compute().ravel()
            W = tod.weight.compute().ravel()

            # stokes_weight = band_tod.dets.stokes_weight()

            # for stokes_index, stokes in enumerate(self.stokes):
            #     pbar.set_postfix(tod=f"{tod_index + 1}/{len(self.tods)}", stokes=stokes)

            #     s = stokes_weight[:, "IQUV".index(stokes)][..., None]

            map_sum += (W * D) @ P
            map_wgt += W @ np.abs(P)

            # del band_tod

            # logger.info(f"Ran mapper for band {band.name} in {humanize_time(ttime.monotonic() - band_start_s)}.")

        return {
            "data": (map_sum / map_wgt).reshape(self.map_shape),
            "weight": map_wgt.reshape(self.map_shape),
            "sum": map_sum.reshape(self.map_shape),
        }



from __future__ import annotations

import os

import maria
import matplotlib.pyplot as plt
import pytest
from maria import all_maps
from maria.io import fetch


def test_change_cache_dir():
    try:
        maria.set_cache_dir("/tmp/maria-data-2")
        fetch("maps/cluster1.fits")
        assert os.path.isfile(os.environ["MARIA_CACHE_DIR"] + "/maps/cluster1.fits")
        maria.set_cache_dir("/tmp/maria-data")
    except Exception:
        del os.environ["MARIA_CACHE_DIR"]
        maria.set_cache_dir("/tmp/maria-data")
        raise Exception()


@pytest.mark.parametrize("filename", all_maps)
def test_all_maps(filename):
    m = maria.map.load(filename=fetch(filename), width=0.1, center=(150, 10))  # noqa
    m.plot()

    plt.close("all")



from __future__ import annotations

import logging
import os
import re

import h5py
import matplotlib as mpl
import numpy as np
import pandas as pd
from astropy.io import fits
from matplotlib.colors import ListedColormap

from ..coords import frames
from ..units import Quantity
from .base import MAP_QUANTITIES, Map, concatenate  # noqa
from .healpix import HEALPixMap  # noqa
from .projection import ProjectionMap  # noqa

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")

with open(f"{here}/maps.txt", "r") as f:
    all_maps = f.read().splitlines()

# from https://gist.github.com/zonca/6515744
cmb_cmap = ListedColormap(
    np.loadtxt(f"{here}/../plotting/Planck_Parchment_RGB.txt") / 255.0,
    name="cmb",
)
cmb_cmap.set_bad("white")
mpl.colormaps.register(cmb_cmap)

MAP_SIZE_KWARGS = ["width", "height", "x_res", "y_res", "resolution"]
VALID_MAP_KWARGS = ["stokes", "nu", "t", "center", "frame", "units", "beam", *MAP_SIZE_KWARGS]

FITS_KEYWORD_MAPPING = {
    "frame": ["FRAME"],
    "units": ["BUNIT", "BUNITS"],
    "nu": ["RESTFREQ"],
}


def load(filename: str, format: str = "auto", **map_kwargs) -> Map:
    if format == "auto":
        format = filename.split(".")[-1].lower()

    if format == "fits":
        data, metadata = read_fits(filename)
    elif format == "h5":
        data, metadata = read_hdf(filename)
    else:
        raise NotImplementedError(f"Unsupported filetype '.{format}'.")

    # if there are any kwargs specifying the size of the map,
    # then remove size kwargs from the read-in map's metadata.
    size_kwargs = [k for k in map_kwargs if k in MAP_SIZE_KWARGS]

    if size_kwargs:
        metadata = {k: v for k, v in metadata.items() if k not in MAP_SIZE_KWARGS}

    metadata.update(map_kwargs)
    logger.debug(f"Loading ProjectionMap with metadata {metadata}")

    return ProjectionMap(data=data, **metadata)


def read_hdf(filename: str):
    with h5py.File(filename, "r") as f:
        data = f["data"][:]

        metadata = {}
        for field in VALID_MAP_KWARGS:
            if field in f.keys():
                value = f[field][()]
                metadata[field] = value if not isinstance(value, bytes) else value.decode()

        if "weight" in f.keys():
            metadata["weight"] = f["weight"][:]

    return data, metadata


def read_fits(filename: str, index: int = 0):
    if not os.path.exists(filename):
        raise FileNotFoundError(filename)

    with fits.open(filename) as hdul:
        indices_with_image = [index for index, h in enumerate(hdul) if h.data is not None]
        if len(indices_with_image) == 0:
            raise ValueError(f"FITS file '{filename}' has no images.")

        index = index or indices_with_image[0]

        hdu = hdul[index]
        header_dict = dict(hdu.header)

        # FITS counts from the bottom, while normal people count from the top
        data = hdu.data[::-1, :]
        if data.ndim < 2:
            raise ValueError("Map should have at least 2 dimensions.")

        frame = {}
        center = {}
        metadata = {}
        beam = [0, 0, 0]
        for axis in range(header_dict["NAXIS"]):
            CTYPE = header_dict.get(f"CTYPE{axis + 1}", None)
            if CTYPE is None:
                logger.debug(f"Could not find CTYPE for dim {dim}")
                break

            if axis < 2:
                dim = "xy"[axis]
                match = re.compile(r"(.+?)-*([A-Z]{3})").match(CTYPE)
                if match:
                    coord, proj = match.groups()
                    frame[dim] = coord
                    logger.debug(f"Using CTYPE{axis + 1} = {CTYPE} for dim {dim}")
                else:
                    raise ValueError(f"Invalid CTYPE {CTYPE}")

                units = header_dict.get(f"CUNIT{axis + 1}", "deg")
                logger.debug(f"Using CUNIT{axis + 1} = {units} for dim {dim}")

                if f"CRVAL{axis + 1}" in header_dict:
                    value = header_dict[f"CRVAL{axis + 1}"]
                    logger.debug(f"Using CRVAL{axis + 1} = {value} for dim {dim}")
                    center[dim] = Quantity(
                        value, units
                    ).deg  # negative because from inside the sphere, latitude goes right to left

                if f"CDELT{axis + 1}" in header_dict:
                    value = header_dict[f"CDELT{axis + 1}"]
                    logger.debug(f"Using CDELT{axis + 1} = {value} for dim {dim}")
                    metadata[f"{dim}_res"] = Quantity(
                        value, units
                    ).deg  # negative because from inside the sphere, latitude goes right to left

            elif CTYPE in ["FREQ", "NU"]:
                freq_units = header_dict.get(f"CUNIT{axis + 1}", "Hz")
                logger.debug(f"Using CUNIT{axis + 1} = {freq_units} for dim {dim}")
                metadata["nu"] = Quantity(
                    header_dict[f"CRVAL{axis + 1}"] + np.arange(header_dict[f"NAXIS{axis + 1}"]), freq_units
                ).Hz
                logger.debug(f"Using CRVAL{axis + 1} = {metadata['nu']} for kwargs nu")

        beam_keys = [["BMAJ", "BMAJOR"], ["BMIN", "BMINOR"], ["BPA"]]
        for index in range(3):
            for key in beam_keys[index]:
                if key in header_dict:
                    value = header_dict[key]
                    beam[index] = Quantity(value, units).deg
                    logger.debug(f"Found beam param {key} = {value}")

        for fits_key in header_dict:
            for maria_key in FITS_KEYWORD_MAPPING:
                for mapped_key in FITS_KEYWORD_MAPPING[maria_key]:
                    if mapped_key == fits_key:
                        value = hdu.header[fits_key]
                        logger.debug(f"Using {fits_key} = {value} for kwarg {maria_key}")
                        metadata[maria_key] = value

        # logger.warning("Could not infer spatial dimensions from FITS header; specify map dimensions manually")

    for frame_name, f in frames.items():
        if (f["FITS_phi"], f["FITS_theta"]) == (frame["x"], frame["y"]):
            metadata["frame"] = frame_name
            logger.debug(f"Using frame '{frame_name}'")

    if "frame" not in metadata:
        metadata["frame"] = "ra/dec"
        logger.warning(f"Could not infer coordinate system from FITS header; assuming frame 'ra/dec'")

    metadata["center"] = (center.get("x"), center.get("y"))
    metadata["degrees"] = "deg" in units
    metadata["beam"] = beam

    if metadata["frame"] in ["ra/dec", "galactic"]:
        data = data[..., ::-1]

    return data, metadata



from __future__ import annotations

import maria
import matplotlib.pyplot as plt
from maria.instrument import Band
from maria.io import fetch
from maria.mappers import MaximumLikelihoodMapper


def test_ml_mapper():
    map_filename = fetch("maps/cluster1.fits", refresh=True)

    f090 = Band(center=90e9, width=20e9, NET_RJ=5e-5)
    f150 = Band(center=150e9, width=30e9, NET_RJ=5e-5)

    array = {"field_of_view": 0.02, "bands": [f090, f150], "primary_size": 50}

    instrument = maria.get_instrument(array=array)

    map_filename = fetch("maps/cluster1.fits")

    input_map = maria.map.load(filename=map_filename, nu=150e9, width=0.1, center=(150, 10))

    input_map.data *= 1e3

    planner = maria.Planner(target=input_map, site="cerro_toco", constraints={"el": (45, 90)})

    plans = planner.generate_plans(
        total_duration=60, sample_rate=50, scan_pattern="daisy", scan_options={"radius": input_map.width.deg / 3}
    )

    sim = maria.Simulation(
        instrument,
        plans=plans,
        site="llano_de_chajnantor",
        map=input_map,
        atmosphere="2d",
    )

    tods = sim.run()

    mapper = MaximumLikelihoodMapper(
        tods=tods,
    )

    mapper.fit()
    mapper.map.to("Jy/beam").plot()

    plt.close("all")



from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Union

import arrow
import jax.numpy as jnp
import numpy as np
import pandas as pd
import scipy as sp
from arrow import Arrow
from astropy.coordinates import EarthLocation
from astropy.io import fits
from astropy.wcs import WCS
from matplotlib import pyplot as plt

from .. import coords
from ..coords import Coordinates, frames
from ..io import DEFAULT_TIME_FORMAT, repr_lat_lon, repr_phi_theta
from ..units import Quantity
from ..utils import compute_diameter, read_yaml
from .patterns import get_scan_pattern_generator, scan_patterns
from .plan import Plan
from .plan_list import PlanList  # noqa
from .planner import Planner  # noqa

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

all_patterns = list(scan_patterns.index.values)

MAX_VELOCITY_WARN = 10  # in deg/s
MAX_ACCELERATION_WARN = 10  # in deg/s

MIN_ELEVATION_WARN = 20  # in deg
MIN_ELEVATION_ERROR = 10  # in deg

here, this_filename = os.path.split(__file__)

PLAN_CONFIGS = {}
for plans_path in Path(f"{here}/plans").glob("*.yml"):
    PLAN_CONFIGS.update(read_yaml(plans_path))

plan_params = set()
for key, config in PLAN_CONFIGS.items():
    plan_params |= set(config.keys())
plan_data = pd.DataFrame(PLAN_CONFIGS).T
all_plans = list(plan_data.index.values)


class UnsupportedPlanError(Exception):
    def __init__(self, invalid_plan):
        super().__init__(
            f"The plan '{invalid_plan}' is not a supported plan. Supported plans are: \n\n{plan_data.to_string()}",
        )


def get_plan_config(plan_name="ten_second_zenith_stare", **kwargs):
    if plan_name not in PLAN_CONFIGS.keys():
        raise UnsupportedPlanError(plan_name)
    plan_config = PLAN_CONFIGS[plan_name].copy()
    for k, v in kwargs.items():
        plan_config[k] = v
    return plan_config


def get_plan(plan_name="ten_second_zenith_stare", **kwargs):
    plan_config = get_plan_config(plan_name, **kwargs)
    return Plan.generate(**plan_config)


PLAN_FIELDS = {
    "start_time": Union[float, str, Arrow],
    "duration": float,
    "sample_rate": float,
    "frame": str,
    "degrees": bool,
    "scan_center": float,
    "scan_pattern": float,
    "scan_options": Mapping,
}


def validate_pointing_kwargs(kwargs):
    """
    Make sure that we have all the ingredients to produce the plan data.
    """
    if ("end_time" not in kwargs.keys()) and ("duration" not in kwargs.keys()):
        raise ValueError(
            """One of 'end_time' or 'duration' must be in the plan kwargs.""",
        )



import logging
import os
import re
from functools import cached_property

import numpy as np
import pandas as pd
import yaml

from ..utils import deg_to_signed_dms, deg_to_signed_hms
from .prefixes import PREFIXES

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

PREFIXES = pd.read_csv(f"{here}/prefixes.csv", index_col=0)
PREFIXES.loc[""] = "", "", "", 0, 1e0
PREFIXES.sort_values("factor", ascending=True, inplace=True)
PREFIXES.loc[:, "primary"] = np.log10(PREFIXES.factor.values) % 3 == 0

dim_entries = {}
quantity_entries = {}
with open(f"{here}/quantities.yml") as f:
    unit_entries = {}
    for quantity, q in yaml.safe_load(f).items():
        q_units = q.pop("units")
        for unit, config in q_units.items():
            config["quantity"] = quantity
            config["aliases"] = set([unit, config["long_name"], *config.get("aliases", [])])
        unit_entries.update(q_units)
        dim_entries[quantity] = q["recipe"]
        quantity_entries[quantity] = q

QUANTITIES = pd.DataFrame(quantity_entries).T
DIMENSIONS = pd.DataFrame(dim_entries).sort_index().T.fillna(0)
UNITS = pd.DataFrame(unit_entries).T
UNITS["human"] = UNITS["human"].astype(bool).fillna(True)
UNITS["symbol"] = UNITS["symbol"].fillna("")
UNITS["factor"] = UNITS["factor"].astype(float)

prefixes_phrase = r"|".join(PREFIXES.index)
base_units_phrase = r"|".join([alias for _, entry in UNITS.iterrows() for alias in entry.aliases]).replace("^", "\\^")
units_pattern = re.compile(rf"^(?P<prefix>({prefixes_phrase}))(?P<raw_unit>{base_units_phrase})(\^(?P<power>.*))?$")  # noqa


def parse_units(units):
    match = units_pattern.search(units)
    if match is None:
        raise ValueError(
            f"Invalid units '{units}'. Valid units are a combination of an SI prefix "
            f"(one of {prefixes_phrase}) and a base unit (one of {base_units_phrase}).",
        )

    u = match.groupdict()
    for _, entry in UNITS.iterrows():
        if u["raw_unit"] in entry.aliases:
            u["raw_unit"] = entry.name
            break

    u.update(UNITS.loc[u["raw_unit"]])
    prefix = PREFIXES.loc[u["prefix"]]
    u["units"] = f"{prefix.name}{u['raw_unit']}"
    u["long_name"] = f"{prefix.long_name}{u['long_name']}"
    u["math_name"] = f"{prefix.symbol_latex}{u['math_name']}"
    u["factor"] *= prefix.factor
    if not u["symbol"]:
        u.pop("symbol")
    return u


def get_factor_and_base_units_vector(units):
    factor = 1
    base_units_vector = np.zeros(len(DIMENSIONS.columns))

    for subunit in units.split(" "):
        u = parse_units(subunit)
        u.update(UNITS.loc[u["raw_unit"]])
        u["power"] = float(u["power"]) if u["power"] else 1.0
        base_units_vector += u["power"] * np.array(DIMENSIONS.loc[u["quantity"]])
        factor *= (u["factor"] * PREFIXES.loc[u["prefix"], "factor"]) ** u["power"]

    return factor, pd.Series(base_units_vector, index=DIMENSIONS.columns)


def repr_power(thing, power):
    if power == 0:
        return None
    if power == 1:
        pow_string = ""
    elif power == int(power):
        pow_string = f"^{int(power)}"
    else:
        pow_string = f"^{power}"
    return f"{thing}{pow_string}"


class Quantity:
    def __new__(cls, value: float, units: str | pd.Series):
        self = super().__new__(cls)

        if isinstance(value, Quantity):
            self.base_units_value = value.base_units_value
            self.base_units_vector = value.base_units_vector

        else:
            if isinstance(units, pd.Series):
                self.base_units_value = np.array(value).astype(float)
                self.base_units_vector = units

            else:
                factor, base_units_vector = get_factor_and_base_units_vector(units)
                if np.ndim(value) == 1:
                    unique_quantities = np.unique([x.quantity for x in value if isinstance(x, Quantity)])
                    if len(unique_quantities) > 1:
                        raise ValueError("'value' contains Quantities of more than one dimension vector")
                    self.base_units_value = factor * np.array([x.to(units) if isinstance(x, Quantity) else x for x in value])
                    self.base_units_vector = base_units_vector

                elif isinstance(units, str):
                    self.base_units_value = factor * np.array(value).astype(float)
                    self.base_units_vector = base_units_vector

                else:
                    raise ValueError("'units' must be a string")

        if (self.base_units_vector == 0).all():
            return self.base_units_value

        return self

    @cached_property
    def quantity(self):
        return self.u["quantity"]

    @cached_property
    def q(self):
        return QUANTITIES.loc[self.quantity].to_dict()

    @cached_property
    def u(self):
        if not self.composite():
            return parse_units(self.units)
        return {
            "prefix": "",
            "raw_units": self.base_units,
            "power": None,
            "factor": 1e0,
            "long_name": "jansky",
            "math_name": f"\\text{{{self.base_units}}}",
            "min_prefix_power": 0,
            "max_prefix_power": 0,
            "human": True,
            "quantity": self.base_units,
            "aliases": {},
            "symbol": "",
            "units": self.base_units,
        }

    def to(self, units):
        factor, base_units_vector = get_factor_and_base_units_vector(units)
        if (base_units_vector == self.base_units_vector).all():
            return self.base_units_value / factor
        else:
            raise ValueError()

    def composite(self):
        quantity_units = self.known_compatible_units()
        return len(quantity_units) == 0

    @cached_property
    def quantity(self):
        qmask = (self.base_units_vector == DIMENSIONS).all(axis=1)
        if qmask.sum() > 0:
            return DIMENSIONS.loc[qmask].iloc[0].name
        else:
            return self.base_units

    def known_compatible_units(self):
        return UNITS.loc[UNITS.quantity == self.quantity]

    @cached_property
    def base_units(self):
        if not hasattr(self, "_base_units"):
            unit_parts = []
            for unit, power in self.base_units_vector.items():
                part = repr_power(unit, power)
                if part is not None:
                    unit_parts.append(part)
            self._base_units = " ".join(unit_parts)
        return self._base_units

    def pin(self, units, inplace=False):
        if inplace:
            self.pinned_units = units
        else:
            pinned_quantity = type(self)(self.base_units_value, self.base_units)
            pinned_quantity.pin(units, inplace=True)
            return pinned_quantity

    @cached_property
    def value(self):
        return self.to(self.units)

    @cached_property
    def units(self):
        if hasattr(self, "pinned_units"):
            return self.pinned_units

        abs_value = np.abs(self.base_units_value)
        abs_finite_value = np.where((abs_value > 0) & np.isfinite(abs_value), abs_value, np.nan)

        if (abs_finite_value > 0).any():
            fid_x = 2 * np.nanquantile(abs_finite_value, q=0.99)

            # all the named units that work for these dimensions
            raw_quantity_units = self.known_compatible_units()
            raw_quantity_units = raw_quantity_units.loc[raw_quantity_units.human]

            if len(raw_quantity_units) > 0:
                prefixed_quantity_units = pd.DataFrame(columns=["value"])
                for unit, unit_entry in raw_quantity_units.iterrows():
                    prefix_mask = (
                        (PREFIXES.power % 3 == 0)
                        & (unit_entry.min_prefix_power <= PREFIXES.power)
                        & (PREFIXES.power <= unit_entry.max_prefix_power)
                    )
                    for prefix, prefix_entry in PREFIXES.loc[prefix_mask].iterrows():
                        prefixed_quantity_units.loc[f"{prefix}{unit}"] = fid_x / (prefix_entry.factor * unit_entry.factor)

                ideal_mask = (prefixed_quantity_units.value > 0.5) & (prefixed_quantity_units.value < 500)
                if ideal_mask.sum() > 0:
                    repr_units = prefixed_quantity_units.loc[ideal_mask].sort_values("value").iloc[0].name
                else:
                    score = np.abs(np.log(prefixed_quantity_units.value / 16))
                    repr_units = prefixed_quantity_units.iloc[score.argmin()].name
                return repr_units

        return self.base_units

    def __repr__(self) -> str:
        if not hasattr(self, "value"):
            if not self.composite():
                self.units = self.compute_human_units()
                self.value = self.to(self.units)
            else:
                self.units = self.base_units
                self.value = self.base_units_value

        if self.quantity == "time":
            if np.any(self.s > 3600):
                return self.timestring
        value_repr = f"{self.value:.04g}" if np.isscalar(self.value) else self.value
        return f"{value_repr}{UNITS['symbol'].get(self.units) or f' {self.units}'}"

    def __neg__(self):
        return type(self)(-self.base_units_value, units=self.base_units_vector)

    def __add__(self, other):
        if (self.base_units_vector == other.base_units_vector).all():
            return type(self)(self.base_units_value + other.base_units_value, units=self.base_units_vector)
        else:
            raise RuntimeError(f"Cannot add units {self} and {other}")

    def __sub__(self, other):
        return self + -other

    def __mul__(self, other):
        if isinstance(other, Quantity):
            return type(self)(
                self.base_units_value * other.base_units_value, units=self.base_units_vector + other.base_units_vector
            )
        else:
            return type(self)(self.base_units_value * other, units=self.base_units_vector)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Quantity):
            return type(self)(
                self.base_units_value / other.base_units_value, units=self.base_units_vector - other.base_units_vector
            )
        else:
            return type(self)(self.base_units_value / other, units=self.base_units_vector)

    def __rtruediv__(self, other):
        return other * (self**-1)

    def __pow__(self, power):
        return type(self)(self.base_units_value**power, units=self.base_units_vector * power)

    # def __iter__(self):
    #     u = self.units
    #     for x in range(self.value):
    #         yield Quantity(x, units=u)

    def __copy__(self):
        return Quantity(self.base_units_value, self.base_units_vector)

    def __deepcopy__(self, memo):
        return self.__copy__()

    @property
    def timestring(self) -> str:
        if self.quantity != "time":
            raise ValueError("'timestring' is only for times")
        parts = []
        t = self.seconds
        for k, v in {"y": 365 * 86400, "d": 86400, "h": 3600, "m": 60}.items():
            if t > v:
                parts.append(f"{int(t // v)}{k}")
                t = t % v
        parts.append(f"{t:.03f}s")
        return " ".join(parts)

    def repr(self, format: str) -> str:
        if format == "dms":
            if self.quantity != "angle":
                raise ValueError("string format 'dms' is only for angles")
            sign, d, m, s = deg_to_signed_dms(self.deg)
            return f"{int(sign * d):>02}°{int(m):>02}’{s:.02f}”"

        if format == "hms":
            if self.quantity != "angle":
                raise ValueError("string format 'hms' is only for angles")
            sign, h, m, s = deg_to_signed_hms(self.deg)
            return f"{int(sign * h):>02}ʰ{int(m):>02}ᵐ{s:.02f}ˢ"

        if format == "deg":
            if self.quantity != "angle":
                raise ValueError("string format 'deg' is only for angles")
            return f"{self.deg:.04f}°"

    def mean(self, axis=None, *args, **kwargs):
        return Quantity(np.mean(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    def median(self, axis=None, *args, **kwargs):
        return Quantity(np.median(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    def min(self, axis=None, *args, **kwargs):
        return Quantity(np.min(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    def max(self, axis=None, *args, **kwargs):
        return Quantity(np.max(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    def std(self, axis=None, *args, **kwargs):
        return Quantity(np.std(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    def ptp(self, axis=None, *args, **kwargs):
        return Quantity(np.ptp(self.base_units_value, axis=axis, *args, **kwargs), units=self.base_units)

    @property
    def shape(self):
        return np.shape(self.base_units_value)

    @property
    def size(self):
        return np.size(self.base_units_value)

    @property
    def ndim(self):
        return np.ndim(self.base_units_value)

    def __getattr__(self, attr):
        try:
            return self.to(attr)
        except Exception:
            pass
        raise AttributeError(f"'Quantity' object has no attribute '{attr}'")

    def __getitem__(self, key):
        return Quantity(self.base_units_value[key], units=self.base_units)

    def __format__(self, *args, **kwargs):
        return repr(self).__format__(*args, **kwargs)

    # def __float__(self):
    #     return float(self.base_units_value)

    def __bool__(self):
        return self.size > 0

    def __len__(self):
        return len(self.base_units_value)

    def __neg__(self):
        return Quantity(-self.base_units_value, units=self.base_units)

    def convert_other(self, other):
        if isinstance(other, Quantity):
            if (self.base_units_vector == other.base_units_vector).all():
                return other.to(self.base_units)
            else:
                raise TypeError(f"Cannot combine quantities '{self.quantity}' to quantity '{other.quantity}'")
        elif np.all(other == 0):
            return other
        raise TypeError(f"{self} and {other} are incompatible quantities")

    def __eq__(self, other):
        return self.base_units_value == self.convert_other(other)

    def __lt__(self, other):
        return self.base_units_value < self.convert_other(other)

    def __gt__(self, other):
        return self.base_units_value > self.convert_other(other)

    def __le__(self, other):
        return self.base_units_value <= self.convert_other(other)

    def __ge__(self, other):
        return self.base_units_value >= self.convert_other(other)
